"""
Integration tests that use the command-line entry points run_whatshap, run_haplotag etc.
"""
import os
from collections import namedtuple

from pytest import raises, fixture, mark
import pysam
from whatshap.cli.phase import run_whatshap
from whatshap.cli import CommandLineError
from whatshap.vcf import VcfReader, VariantCallPhase


trio_bamfile = "tests/data/trio.pacbio.bam"
trio_merged_bamfile = "tests/data/trio-merged-blocks.bam"
trio_paired_end_bamfile = "tests/data/paired_end.sorted.bam"
ped_samples_bamfile = "tests/data/ped_samples.bam"
recombination_breaks_bamfile = "tests/data/recombination_breaks.sorted.bam"
quartet2_bamfile = "tests/data/quartet2.bam"
short_bamfile = "tests/data/short-genome/short.bam"
short_duplicate_bamfile = "tests/data/short-genome/short-one-read-duplicate.bam"
indels_bamfile = "tests/data/indels.bam"
dist_geno_bamfile = "tests/data/test_dist_geno.bam"

bam_files = [
    trio_bamfile,
    trio_merged_bamfile,
    trio_paired_end_bamfile,
    recombination_breaks_bamfile,
    quartet2_bamfile,
    short_bamfile,
    short_duplicate_bamfile,
    indels_bamfile,
    dist_geno_bamfile,
]


@fixture(params=["whatshap", "hapchat"])
def algorithm(request):
    return request.param


def setup_module():
    # This function is run once for this module
    for bam_path in bam_files:
        assert bam_path.endswith(".bam")
        sam_path = bam_path[:-4] + ".sam"
        pysam.view(sam_path, "-b", "-o", bam_path, catch_stdout=False)
        pysam.index(bam_path, catch_stdout=False)


def teardown_module():
    for path in bam_files:
        os.remove(path)
        os.remove(path + ".bai")


def test_run_phase_without_reference():
    from whatshap.__main__ import main

    with raises(SystemExit):
        main(["phase", "-o", "/dev/null", "tests/data/onevariant.vcf", "tests/data/oneread.bam"])


def test_one_variant(algorithm):
    run_whatshap(
        phase_input_files=["tests/data/oneread.bam"],
        variant_file="tests/data/onevariant.vcf",
        output="/dev/null",
        algorithm=algorithm,
    )


def test_default_output(algorithm):
    """Output to stdout"""
    run_whatshap(
        phase_input_files=["tests/data/oneread.bam"],
        variant_file="tests/data/onevariant.vcf",
        algorithm=algorithm,
    )


def test_one_variant_cram(algorithm):
    run_whatshap(
        phase_input_files=["tests/data/oneread.cram"],
        reference="tests/data/oneread-ref.fasta",
        variant_file="tests/data/onevariant.vcf",
        output="/dev/null",
        algorithm=algorithm,
    )


def test_cram_no_reference(algorithm):
    # This needs to fail because CRAM requires a reference, but it was not given.

    # If REF_PATH is not set, pysam/htslib tries to retrieve the reference from EBI via
    # the internet.
    os.environ["REF_PATH"] = "/does/not/exist"
    with raises(CommandLineError):
        run_whatshap(
            phase_input_files=["tests/data/oneread.cram"],
            variant_file="tests/data/onevariant.vcf",
            output="/dev/null",
            algorithm=algorithm,
        )


def test_bam_without_readgroup(algorithm):
    run_whatshap(
        phase_input_files=["tests/data/no-readgroup.bam"],
        variant_file="tests/data/onevariant.vcf",
        output="/dev/null",
        ignore_read_groups=True,
        algorithm=algorithm,
    )


def test_requested_sample_not_found(algorithm):
    with raises(CommandLineError):
        run_whatshap(
            phase_input_files=["tests/data/oneread.bam"],
            variant_file="tests/data/onevariant.vcf",
            output="/dev/null",
            samples=["DOES_NOT_EXIST"],
            algorithm=algorithm,
        )


@mark.parametrize(
    "algorithm,expected_vcf",
    [
        ("whatshap", "tests/data/pacbio/phased.vcf"),
        ("hapchat", "tests/data/pacbio/phased_hapchat.vcf"),
    ],
)
def test_with_reference(algorithm, expected_vcf, tmpdir):
    # This tests also whether lowercase reference FASTA files work:
    # If lowercase and uppercase are treated differently, then the
    # output is slightly different from the expected.

    # note: because hapchat has a different dynamic programming
    # scheme, it may phase some variants differently, e.g., the
    # variant at site 11221 of phased.vcf.  It also phases each
    # heterozygous site, even if the scores (in the DP table) of
    # its (two) possible phasings are identical -- such is the
    # case for sites 13300 and 14324 of phased.vcf.  It is for
    # this reason that we have a second phased_hapchat.vcf which
    # is different in these above three sites.  Whether or not
    # this a desired behaviour is subject to discussion --
    # possible handling (i.e., avoiding the phasing of) sites with
    # identical phasing scores is a possible future work, etc.
    out = str(tmpdir.join("out.vcf"))
    run_whatshap(
        phase_input_files=["tests/data/pacbio/pacbio.bam"],
        variant_file="tests/data/pacbio/variants.vcf",
        reference="tests/data/pacbio/reference.fasta",
        output=out,
        write_command_line_header=False,  # for easier VCF comparison
        algorithm=algorithm,
    )
    print("out:", out)
    with open(expected_vcf) as f:
        expected = f.read()
    with open(out) as f:
        actual = f.read()

    assert actual == expected, "VCF output not as expected"


def test_with_reference_and_indels(algorithm):
    run_whatshap(
        phase_input_files=["tests/data/pacbio/pacbio.bam"],
        variant_file="tests/data/pacbio/variants.vcf",
        reference="tests/data/pacbio/reference.fasta",
        indels=True,
        algorithm=algorithm,
    )


@mark.parametrize(
    "algorithm,expected_lines",
    [
        (
            "whatshap",
            [
                "1\t60906167\t.\tG\tA\t.\tPASS\tAC=2;AN=6\tGT:PS\t0/1:.\t0|1:60906167\t0/0:.\n",
                "1\t60907394\t.\tG\tA\t.\tPASS\tAC=4;AN=6\tGT:PS\t0|1:60907394\t1/1:.\t0/1:.\n",
                "1\t60907460\t.\tG\tT\t.\tPASS\tAC=2;AN=6\tGT:PS\t0|1:60907394\t0|1:60906167\t0/0:.\n",
                "1\t60907473\t.\tC\tA\t.\tPASS\tAC=2;AN=6\tGT:PS\t0|1:60907394\t0/1:.\t0/0:.\n",
                "1\t60909718\t.\tT\tC\t.\tPASS\tAC=2;AN=6\tGT\t0/1\t0/1\t0/0\n",
            ],
        ),
        (
            "hapchat",
            [
                "1\t60906167\t.\tG\tA\t.\tPASS\tAC=2;AN=6\tGT:PS\t0/1:.\t1|0:60906167\t0/0:.\n",
                "1\t60907394\t.\tG\tA\t.\tPASS\tAC=4;AN=6\tGT:PS\t1|0:60907394\t1/1:.\t0/1:.\n",
                "1\t60907460\t.\tG\tT\t.\tPASS\tAC=2;AN=6\tGT:PS\t1|0:60907394\t1|0:60906167\t0/0:.\n",
                "1\t60907473\t.\tC\tA\t.\tPASS\tAC=2;AN=6\tGT:PS\t1|0:60907394\t0/1:.\t0/0:.\n",
                "1\t60909718\t.\tT\tC\t.\tPASS\tAC=2;AN=6\tGT\t0/1\t0/1\t0/0\n",
            ],
        ),
    ],
)
def test_ps_tag(algorithm, expected_lines, tmpdir):
    out = str(tmpdir.join("out.vcf"))
    run_whatshap(
        variant_file="tests/data/trio.vcf",
        phase_input_files=["tests/data/trio.pacbio.bam"],
        output=out,
        tag="PS",
        algorithm=algorithm,
    )
    with open(out) as f:
        lines = [line for line in f.readlines() if not line.startswith("#")]

    # TODO This is quite an ugly way to test phased VCF writing (see parametrization)
    for i in range(5):
        assert lines[i] == expected_lines[i]


# def assert_phasing(phases, expected_phases):
#    # TODO: this code is not "block aware". Would be useful to extend it to compare phasings per block
#    print('assert_phasing({}, {})'.format(phases, expected_phases))
#    assert len(phases) == len(expected_phases)
#    p_unchanged = []
#    p_inverted = []
#    p_expected = []
#    for phase, expected_phase in zip(phases, expected_phases):
#        if (phase is None) and (expected_phase is None):
#            continue
#        assert phase is not None and expected_phase is not None
#        assert phase.block_id == expected_phase.block_id
#        p_unchanged.append(phase.phase)
#        p_inverted.append(1-phase.phase)
#        p_expected.append(expected_phase.phase)
#    assert (p_unchanged == p_expected) or (p_inverted == p_expected)


def assert_phasing(phases, expected_phases):
    # TODO: this code is not "block aware". Would be useful to extend it to compare phasings per block
    print("assert_phasing({}, {})".format(phases, expected_phases))
    assert len(phases) == len(expected_phases)
    haplotypes = []
    expected_haplotypes = []
    for phase, expected_phase in zip(phases, expected_phases):
        if (phase is None) and (expected_phase is None):
            continue
        assert phase is not None and expected_phase is not None
        assert phase.block_id == expected_phase.block_id
        haplotypes.append(phase.phase)
        expected_haplotypes.append(expected_phase.phase)
    # construct haplotype sequences
    n_positions = len(haplotypes)
    if n_positions > 0:
        ploidy = len(haplotypes[0])
        haplotype_sequences = [""] * ploidy
        expected_haplotype_sequences = [""] * ploidy
        for i in range(n_positions):
            for p in range(ploidy):
                haplotype_sequences[p] += str(haplotypes[i][p])
                expected_haplotype_sequences[p] += str(expected_haplotypes[i][p])
        assert sorted(haplotype_sequences) == sorted(expected_haplotype_sequences)


def test_phase_three_individuals(algorithm, tmpdir):
    outvcf = str(tmpdir.join("output.vcf"))
    outreadlist = str(tmpdir.join("readlist.tsv"))
    run_whatshap(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio.vcf",
        read_list_filename=outreadlist,
        output=outvcf,
        algorithm=algorithm,
    )
    assert os.path.isfile(outvcf)
    assert os.path.isfile(outreadlist)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 5
    assert table.samples == ["HG004", "HG003", "HG002"]

    phase1 = VariantCallPhase(60906167, (0, 1), None)
    phase3 = VariantCallPhase(60907394, (0, 1), None)
    assert_phasing(table.phases_of("HG004"), [None, phase3, phase3, phase3, None])
    assert_phasing(table.phases_of("HG003"), [phase1, None, phase1, None, None])
    assert_phasing(table.phases_of("HG002"), [None, None, None, None, None])


def test_phase_one_of_three_individuals(algorithm, tmpdir):
    outvcf = str(tmpdir.join("output.vcf"))
    run_whatshap(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio.vcf",
        output=outvcf,
        samples=["HG003"],
        algorithm=algorithm,
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 5
    assert table.samples == ["HG004", "HG003", "HG002"]

    phase0 = VariantCallPhase(60906167, (0, 1), None)
    assert_phasing(table.phases_of("HG004"), [None, None, None, None, None])
    assert_phasing(table.phases_of("HG003"), [phase0, None, phase0, None, None])
    assert_phasing(table.phases_of("HG002"), [None, None, None, None, None])


def test_phase_with_phased_blocks(tmp_path):
    outvcf1 = tmp_path / "output1.vcf"
    outvcf2 = tmp_path / "output2.vcf"
    # run whatshap without --ignore-read-groups option
    run_whatshap(
        phase_input_files=[
            "tests/data/phased-blocks.reads.bam",
            "tests/data/phased-blocks.blocks.vcf",
        ],
        variant_file="tests/data/phased-blocks.variants.vcf",
        output=outvcf1,
    )
    # run whatshap with --ignore-read-groups option
    run_whatshap(
        phase_input_files=[
            "tests/data/phased-blocks.reads.bam",
            "tests/data/phased-blocks.blocks.vcf",
        ],
        variant_file="tests/data/phased-blocks.variants.vcf",
        output=outvcf2,
        ignore_read_groups=True,
    )
    # the results should be identical
    with open(outvcf1) as f:
        lines1 = [line for line in f if line[0] != "#"]
    with open(outvcf2) as f:
        lines2 = [line for line in f if line[0] != "#"]

    for l1, l2 in zip(lines1, lines2):
        assert l1 == l2


def test_phase_trio(tmpdir):
    outvcf = str(tmpdir.join("output.vcf"))
    outreadlist = str(tmpdir.join("readlist.tsv"))
    run_whatshap(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio.vcf",
        read_list_filename=outreadlist,
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
    )
    assert os.path.isfile(outvcf)
    assert os.path.isfile(outreadlist)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 5
    assert table.samples == ["HG004", "HG003", "HG002"]

    phase0 = VariantCallPhase(60906167, (0, 1), None)
    assert_phasing(table.phases_of("HG004"), [phase0, phase0, phase0, phase0, phase0])
    assert_phasing(table.phases_of("HG003"), [phase0, None, phase0, phase0, phase0])
    assert_phasing(table.phases_of("HG002"), [None, phase0, None, None, None])


def test_phase_trio_hapchat():
    with raises(CommandLineError) as e:
        run_whatshap(
            phase_input_files=[trio_bamfile],
            variant_file="tests/data/trio.vcf",
            output="/dev/null",
            ped="tests/data/trio.ped",
            algorithm="hapchat",
        )
    assert "cannot do pedigree phasing" in e.value.args[0]


@mark.parametrize("ped_samples", [True, False])
def test_phase_trio_use_ped_samples(ped_samples, tmpdir):
    outvcf = str(tmpdir.join("output_ped_samples.vcf"))
    outreadlist = str(tmpdir.join("readlist.tsv"))
    run_whatshap(
        phase_input_files=[ped_samples_bamfile],
        variant_file="tests/data/ped_samples.vcf",
        read_list_filename=outreadlist,
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
        use_ped_samples=ped_samples,
    )
    assert os.path.isfile(outvcf)
    assert os.path.isfile(outreadlist)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 5
    assert table.samples == ["HG004", "HG003", "HG002", "orphan"]

    phase0 = VariantCallPhase(60906167, (0, 1), None)
    phase1 = VariantCallPhase(60907394, (0, 1), None)
    assert_phasing(table.phases_of("HG004"), [phase0, phase0, phase0, phase0, phase0])
    assert_phasing(table.phases_of("HG003"), [phase0, None, phase0, phase0, phase0])
    assert_phasing(table.phases_of("HG002"), [None, phase0, None, None, None])

    if ped_samples:
        assert_phasing(table.phases_of("orphan"), [None, None, None, None, None])
    else:
        assert_phasing(table.phases_of("orphan"), [None, phase1, phase1, phase1, None])


@mark.parametrize(
    "sample_set",
    [["HG002"], ["HG003"], ["HG004"], ["HG002", "HG003"], ["HG002", "HG004"], ["HG003", "HG004"]],
)
def test_phase_ped_sample(tmpdir, sample_set):
    # running with --ped and --sample on subset of trio, should give same results as running with only --sample
    # the trio information should be ignored
    outvcf1 = str(tmpdir.join("output1.vcf"))
    outvcf2 = str(tmpdir.join("output2.vcf"))
    run_whatshap(
        phase_input_files=[ped_samples_bamfile],
        variant_file="tests/data/ped_samples.vcf",
        output=outvcf1,
        ped="tests/data/trio.ped",
        samples=sample_set,
    )
    run_whatshap(
        phase_input_files=[ped_samples_bamfile],
        variant_file="tests/data/ped_samples.vcf",
        output=outvcf2,
        samples=sample_set,
    )

    assert os.path.isfile(outvcf1)
    assert os.path.isfile(outvcf2)

    tables1 = list(VcfReader(outvcf1, phases=True))
    tables2 = list(VcfReader(outvcf2, phases=True))

    assert len(tables1) == 1 and len(tables2) == 1
    table1, table2 = tables1[0], tables2[0]

    for individual in sample_set:
        assert_phasing(table1.phases_of(individual), table2.phases_of(individual))


def test_phase_trio_distrust_genotypes(tmpdir):
    outvcf = str(tmpdir.join("output_gl.vcf"))
    outreadlist = str(tmpdir.join("readlist.tsv"))
    run_whatshap(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio_genotype_likelihoods.vcf",
        read_list_filename=outreadlist,
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
        distrust_genotypes=True,
    )
    assert os.path.isfile(outvcf)
    assert os.path.isfile(outreadlist)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 5
    assert table.samples == ["HG004", "HG003", "HG002"]

    phase0 = VariantCallPhase(60906167, (0, 1), None)
    assert_phasing(table.phases_of("HG004"), [None, phase0, phase0, phase0, None])
    assert_phasing(table.phases_of("HG003"), [phase0, None, phase0, phase0, phase0])
    assert_phasing(table.phases_of("HG002"), [phase0, None, phase0, phase0, phase0])


def test_phase_trio_merged_blocks(tmpdir):
    outvcf = str(tmpdir.join("output-merged-blocks.vcf"))
    run_whatshap(
        phase_input_files=[trio_merged_bamfile],
        variant_file="tests/data/trio-merged-blocks.vcf",
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 8
    assert table.samples == ["HG002", "HG003", "HG004"]
    assert table.num_of_blocks_of("HG004") == 1
    assert table.num_of_blocks_of("HG003") == 1
    assert table.num_of_blocks_of("HG002") == 1

    phase0 = VariantCallPhase(752566, (0, 1), None)
    phase1 = VariantCallPhase(752566, (1, 0), None)
    assert_phasing(
        table.phases_of("HG004"), [phase1, phase1, phase1, None, phase1, phase1, phase1, phase1]
    )
    assert_phasing(
        table.phases_of("HG003"), [None, None, None, None, phase0, phase0, phase0, phase1]
    )
    assert_phasing(table.phases_of("HG002"), [None, None, None, None, None, None, None, phase1])


def test_phase_trio_dont_merge_blocks(tmpdir):
    outvcf = str(tmpdir.join("output-merged-blocks.vcf"))
    run_whatshap(
        phase_input_files=[trio_merged_bamfile],
        variant_file="tests/data/trio-merged-blocks.vcf",
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
        genetic_haplotyping=False,
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 8
    assert table.samples == ["HG002", "HG003", "HG004"]
    assert table.num_of_blocks_of("HG004") == 2
    assert table.num_of_blocks_of("HG003") == 1
    assert table.num_of_blocks_of("HG002") == 1

    phase1 = VariantCallPhase(752566, (1, 0), None)
    phase2_0 = VariantCallPhase(853954, (0, 1), None)
    phase2_1 = VariantCallPhase(853954, (1, 0), None)
    assert_phasing(
        table.phases_of("HG004"),
        [phase1, phase1, phase1, None, phase2_1, phase2_1, phase2_1, phase2_1],
    )
    assert_phasing(
        table.phases_of("HG003"), [None, None, None, None, phase2_0, phase2_0, phase2_0, phase2_1]
    )
    assert_phasing(table.phases_of("HG002"), [None, None, None, None, None, None, None, phase2_1])


def test_genetic_phasing_symbolic_alt(tmpdir):
    outvcf = str(tmpdir.join("output.vcf"))
    run_whatshap(
        phase_input_files=[],
        variant_file="tests/data/trio-symbolic-alt.vcf",
        output=outvcf,
        ped="tests/data/trio.ped",
        indels=True,
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True, indels=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 5
    assert table.samples == ["HG004", "HG003", "HG002"]

    phase0 = VariantCallPhase(60906167, (0, 1), None)
    assert_phasing(table.phases_of("HG004"), [phase0, phase0, phase0, phase0, phase0])
    assert_phasing(table.phases_of("HG003"), [phase0, None, phase0, phase0, phase0])
    assert_phasing(table.phases_of("HG002"), [None, phase0, None, None, None])


def test_phase_mendelian_conflict(tmpdir):
    outvcf = str(tmpdir.join("output.vcf"))
    run_whatshap(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio-mendelian-conflict.vcf",
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 5
    assert table.samples == ["HG004", "HG003", "HG002"]

    phase = VariantCallPhase(60906167, (0, 1), None)
    assert_phasing(table.phases_of("HG004"), [phase, None, phase, phase, phase])
    assert_phasing(table.phases_of("HG003"), [phase, None, phase, phase, phase])
    assert_phasing(table.phases_of("HG002"), [None, None, None, None, None])


def test_phase_missing_genotypes(tmp_path):
    outvcf = tmp_path / "output.vcf"
    run_whatshap(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio-missing-genotypes.vcf",
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 5
    assert table.samples == ["HG004", "HG003", "HG002"]

    phase = VariantCallPhase(60906167, (0, 1), None)
    assert_phasing(table.phases_of("HG004"), [phase, phase, None, phase, None])
    assert_phasing(table.phases_of("HG003"), [phase, None, None, phase, None])
    assert_phasing(table.phases_of("HG002"), [None, phase, None, None, None])


@mark.parametrize("chromosome", ["1", "2"])
def test_phase_specific_chromosome(chromosome, tmp_path):
    outvcf = tmp_path / "output.vcf"
    run_whatshap(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio-two-chromosomes.vcf",
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
        chromosomes=[chromosome],
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 2
    for table in tables:
        assert len(table.variants) == 5
        assert table.samples == ["HG004", "HG003", "HG002"]
        if table.chromosome == "1" == chromosome:
            phase0 = VariantCallPhase(60906167, (0, 1), None)
            assert_phasing(table.phases_of("HG004"), [phase0, phase0, phase0, phase0, phase0])
            assert_phasing(table.phases_of("HG003"), [phase0, None, phase0, phase0, phase0])
            assert_phasing(table.phases_of("HG002"), [None, phase0, None, None, None])
        elif table.chromosome == "2" == chromosome:
            phase0 = VariantCallPhase(60906167, (0, 1), None)
            phase1 = VariantCallPhase(60906167, (1, 0), None)
            assert_phasing(table.phases_of("HG004"), [phase0, None, None, None, phase1])
            assert_phasing(table.phases_of("HG003"), [phase0, None, None, None, None])
            assert_phasing(table.phases_of("HG002"), [None, None, None, None, phase0])
        else:
            assert_phasing(table.phases_of("HG004"), [None, None, None, None, None])
            assert_phasing(table.phases_of("HG003"), [None, None, None, None, None])
            assert_phasing(table.phases_of("HG002"), [None, None, None, None, None])


def test_phase_trio_paired_end_reads(tmp_path):
    outvcf = tmp_path / "output-paired_end.vcf"
    run_whatshap(
        phase_input_files=[trio_paired_end_bamfile],
        variant_file="tests/data/paired_end.sorted.vcf",
        output=outvcf,
        ped="tests/data/trio_paired_end.ped",
        genmap="tests/data/trio.map",
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 3
    assert table.samples == ["mother", "father", "child"]
    assert table.num_of_blocks_of("mother") == 1
    assert table.num_of_blocks_of("father") == 0
    assert table.num_of_blocks_of("child") == 1

    phase0 = VariantCallPhase(80050, (0, 1), None)
    phase1 = VariantCallPhase(80050, (1, 0), None)

    assert_phasing(table.phases_of("mother"), [phase1, phase1, phase0])
    assert_phasing(table.phases_of("father"), [None, None, None])
    assert_phasing(table.phases_of("child"), [None, None, phase1])


@mark.parametrize(
    "expect_recombination,parameters",
    [
        (False, {"genmap": "tests/data/recombination_breaks.map"}),
        (True, {"recombrate": 1000000}),
        (False, {"recombrate": 0.0000001}),
    ],
)
def test_phase_quartet_recombination_breakpoints(expect_recombination, parameters, tmp_path):
    outvcf = tmp_path / "output-recombination_breaks.vcf"
    outlist = tmp_path / "output.recomb"
    run_whatshap(
        phase_input_files=[recombination_breaks_bamfile],
        variant_file="tests/data/quartet.vcf.gz",
        output=outvcf,
        ped="tests/data/recombination_breaks.ped",
        recombination_list_filename=outlist,
        **parameters,
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 4
    assert table.samples == ["HG002", "HG005", "HG003", "HG004"]
    assert table.num_of_blocks_of("HG002") == 0
    assert table.num_of_blocks_of("HG005") == 0
    assert table.num_of_blocks_of("HG003") == 1
    assert table.num_of_blocks_of("HG004") == 0

    phase0 = VariantCallPhase(68735304, (0, 1), None)
    phase1 = VariantCallPhase(68735304, (1, 0), None)

    assert_phasing(table.phases_of("HG002"), [None, None, None, None])
    assert_phasing(table.phases_of("HG005"), [None, None, None, None])
    if expect_recombination:
        assert_phasing(table.phases_of("HG003"), [phase0, phase0, None, phase1])
    else:
        assert_phasing(table.phases_of("HG003"), [phase0, phase0, None, phase0])
    assert_phasing(table.phases_of("HG004"), [None, None, None, None])

    lines = outlist.read_text().splitlines()
    if expect_recombination:
        assert len(lines) == 3
        assert lines[1] == "HG002 1 68735433 68738308 0 1 0 0 3"
        assert lines[2] == "HG005 1 68735433 68738308 0 1 0 0 3"
    else:
        assert len(lines) == 1


def test_phase_trio_zero_distance(tmp_path):
    outvcf = tmp_path / "output.vcf"
    run_whatshap(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio.vcf",
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/zero-genetic-distance.map",
    )
    assert os.path.isfile(outvcf)


def test_ignore_read_groups(algorithm):
    run_whatshap(
        variant_file="tests/data/pacbio/variants.vcf",
        phase_input_files=["tests/data/pacbio/pacbio.bam"],
        reference="tests/data/pacbio/reference.fasta",
        ignore_read_groups=True,
        output="/dev/null",
        algorithm=algorithm,
    )


def test_readgroup_without_sample_name(algorithm):
    run_whatshap(
        phase_input_files=["tests/data/oneread-readgroup-without-sample.bam"],
        variant_file="tests/data/onevariant.vcf",
        output="/dev/null",
        ignore_read_groups=True,
        algorithm=algorithm,
    )


def test_genetic_haplotyping(tmp_path):
    outvcf = tmp_path / "output.vcf"
    outrecomb = tmp_path / "utput.recomb"
    run_whatshap(
        variant_file="tests/data/genetic-haplotyping.vcf",
        phase_input_files=[],
        ped="tests/data/genetic-haplotyping.ped",
        output=outvcf,
        recombination_list_filename=outrecomb,
    )
    tables = list(VcfReader(outvcf, phases=True))

    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 3
    assert table.samples == ["sampleA", "sampleB", "sampleC", "sampleD", "sampleE"]
    assert table.num_of_blocks_of("sampleA") == 1
    assert table.num_of_blocks_of("sampleB") == 1
    assert table.num_of_blocks_of("sampleC") == 0
    assert table.num_of_blocks_of("sampleD") == 1
    assert table.num_of_blocks_of("sampleE") == 1

    phase0 = VariantCallPhase(10327, (0, 1), None)
    phase1 = VariantCallPhase(10327, (1, 0), None)

    assert_phasing(table.phases_of("sampleA"), [phase0, phase0, phase1])
    assert_phasing(table.phases_of("sampleB"), [phase0, None, None])
    assert_phasing(table.phases_of("sampleC"), [None, None, None])
    assert_phasing(table.phases_of("sampleD"), [phase0, None, phase1])
    assert_phasing(table.phases_of("sampleE"), [phase0, phase0, None])

    lines = [line.split() for line in outrecomb.read_text().splitlines()]
    assert len(lines) == 2
    Fields = namedtuple("Fields", [f.strip("#\n") for f in lines[0]])
    recomb = Fields(*lines[1])
    print(recomb)
    assert recomb.child_id == "sampleC"
    assert recomb.chromosome == "1"
    assert recomb.position1 == "31295"
    assert recomb.position2 == "102596"

    # assert recomb.transmitted_hap_mother1 != recomb.transmitted_hap_mother2
    # assert recomb.transmitted_hap_father1 == recomb.transmitted_hap_father2


def test_quartet2():
    run_whatshap(
        variant_file="tests/data/quartet2.vcf",
        phase_input_files=[quartet2_bamfile],
        ped="tests/data/quartet2.ped",
        output="/dev/null",
    )


@mark.parametrize(
    "algorithm,expected_blocks",
    [("whatshap", [10, 10, None, 200, 200]), ("hapchat", [10, 10, 10, 10, 10])],
)
def test_phased_blocks(algorithm, expected_blocks, tmp_path):
    # This test involves a simple example on a pair of reads which
    # overlap a single site which is homozygous.  While we are
    # distrusting genotypes AND including homozygous sites, i.e.,
    # we are doing full genotyping, if we were phasing purely from
    # the reads, then whether or not the pair of reads falls on
    # the same or different haplotypes should not matter.  With
    # this in mind, reasoning with genotype likelihoods slighly
    # disfavours a cis-phasing of this pair, which is what
    # whatshap does.  While hapchat, however, does take into
    # account genotype likelihoods, while making the
    # all-heterozygous assumption, hence cis-phasing this pair.

    # Note that taking into account genotype likelihoods is a
    # future work planned for hapchat.  Since the nature of the
    # hapchat DP scheme is such that relaxing the all-heterozygous
    # assumption would be too costly in terms of runtime and
    # memory, a possibility is to (re-) exclude homozygous sites
    # in a preprocessing step based on some threshold on the
    # genotype likelihoods.
    outvcf = tmp_path / "output.vcf"
    run_whatshap(
        phase_input_files=[short_bamfile],
        variant_file="tests/data/short-genome/short.vcf",
        ignore_read_groups=True,
        distrust_genotypes=True,
        include_homozygous=True,
        output=outvcf,
        algorithm=algorithm,
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "chr1"
    assert len(table.variants) == 5
    assert table.samples == ["sample"]

    blocks = [(p.block_id if p is not None else None) for p in table.phases_of("sample")]
    assert blocks == expected_blocks


@mark.parametrize(
    "algorithm,expected_block",
    [("whatshap", [10, 10, None, None, None]), ("hapchat", [10, 10, 10, None, None])],
)
def test_duplicate_read(algorithm, expected_block, tmp_path):
    # This test is very similar to the previous test_phased_blocks
    # test, except that there is just a single read this time,
    # with homozygous site.  Still, since hapchat would rather
    # phase this homozygous site, since the context is full
    # genotyping, it does so, regardless of any genotype
    # likelihood.  See above test for more details.
    outvcf = tmp_path / "output.vcf"
    run_whatshap(
        phase_input_files=[short_duplicate_bamfile],
        variant_file="tests/data/short-genome/short.vcf",
        ignore_read_groups=True,
        distrust_genotypes=True,
        include_homozygous=True,
        output=outvcf,
        algorithm=algorithm,
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "chr1"
    assert len(table.variants) == 5
    assert table.samples == ["sample"]

    blocks = [(p.block_id if p is not None else None) for p in table.phases_of("sample")]
    assert blocks == expected_block


def test_wrong_chromosome(algorithm, tmp_path):
    outvcf = tmp_path / "output.vcf"
    with raises(CommandLineError):
        run_whatshap(
            phase_input_files=[short_bamfile],
            ignore_read_groups=True,
            variant_file="tests/data/short-genome/wrongchromosome.vcf",
            output=outvcf,
            algorithm=algorithm,
        )


def test_indel_phasing(algorithm, tmp_path):
    outvcf = tmp_path / "output.vcf"
    run_whatshap(
        phase_input_files=[indels_bamfile],
        indels=True,
        variant_file="tests/data/indels.vcf",
        reference="tests/data/random0.fasta",
        output=outvcf,
        algorithm=algorithm,
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, indels=True, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "random0"
    assert len(table.variants) == 4
    assert table.samples == ["sample1"]

    phase0 = VariantCallPhase(41, (0, 1), None)
    phase1 = VariantCallPhase(41, (1, 0), None)
    assert_phasing(table.phases_of("sample1"), [phase0, phase1, phase0, phase1])


def test_with_read_merging(algorithm):
    run_whatshap(
        phase_input_files=["tests/data/pacbio/pacbio.bam"],
        variant_file="tests/data/pacbio/variants.vcf",
        reference="tests/data/pacbio/reference.fasta",
        output="/dev/null",
        read_merging=True,
        algorithm=algorithm,
    )


def test_vcf_with_missing_headers(algorithm):
    # Since pysam 0.16, this type of invalid VCF is no longer accepted
    with raises(CommandLineError):
        run_whatshap(
            phase_input_files=["tests/data/oneread.bam"],
            variant_file="tests/data/missing-headers.vcf",
            output="/dev/null",
            algorithm=algorithm,
        )


def test_distrust_genotypes_assertion(tmp_path):
    outvcf = tmp_path / "output.vcf"
    run_whatshap(
        indels=False,
        phase_input_files=[dist_geno_bamfile],
        variant_file="tests/data/test_dist_geno.vcf",
        output=outvcf,
    )
    assert os.path.isfile(outvcf)
    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "chr1"
    phase0 = VariantCallPhase(23824647, (0, 1), None)
    assert_phasing(table.phases_of("NA12878"), [None, phase0, None, phase0])



from whatshap.readselect import readselection
from whatshap.testhelpers import string_to_readset


def test_selection():
    reads = string_to_readset(
        """
      1  1
      00
      0   1
      10  1
      1   1
        11
      0   1
      1    1
    """
    )
    selected_reads = readselection(reads, max_cov=1, preferred_source_ids=None, bridging=False)
    assert selected_reads == set([1, 5])
    selected_reads = readselection(reads, max_cov=2, preferred_source_ids=None, bridging=False)
    assert selected_reads == set([1, 3, 5]), str(selected_reads)
    selected_reads = readselection(reads, max_cov=3, preferred_source_ids=None, bridging=False)
    assert selected_reads == set([1, 3, 5, 7]), str(selected_reads)
    selected_reads = readselection(reads, max_cov=3, preferred_source_ids=None, bridging=True)
    # Here the assert is wrong, because the bridging doesn't come into account , because in the slice_read the selected
    # reads  have already coverage 3 by set ([1,3,5,7]) because first each position has to covered at least once before
    # the bridging starts
    assert selected_reads == set([1, 3, 5, 7]), str(selected_reads)


def test_selection2():
    reads = string_to_readset(
        """
      1111
         111
         1  111
         1     11
        1      11
    """
    )
    selected_reads = readselection(reads, max_cov=4, preferred_source_ids=None, bridging=False)
    assert selected_reads == set([0, 1, 2, 3]), str(selected_reads)


def bridging():
    reads = string_to_readset(
        """
      11
      00
        11
        00
          11
          00
      1    1
    """
    )
    selected_reads = readselection(reads, max_cov=2, preferred_source_ids=None, bridging=False)
    assert selected_reads == set([0, 1, 2, 3, 4, 5])
    selected_reads = readselection(reads, max_cov=2, preferred_source_ids=None, bridging=True)
    # Not sure why 0 is there selected and not 1...
    assert selected_reads == set([0, 3, 5, 6])


# Component comparison does not work
def test_components_of_readselection():
    reads = string_to_readset(
        """
      111
         000
      00
          00
       1   1
    """
    )
    selected_reads = readselection(reads, max_cov=2, preferred_source_ids=None, bridging=False)
    assert selected_reads == set([0, 1, 2, 3]), str(selected_reads)
    #    assert len(set(new_components.values())) == 2
    selected_reads = readselection(reads, max_cov=2, preferred_source_ids=None, bridging=True)
    assert selected_reads == set([0, 1, 4]), str(selected_reads)


#      assert len(set(new_components.values())) == 1


def test_selection_with_preferred_sources():
    readset = string_to_readset(
        """
      1        1
    """,
        source_id=3,
    )
    more_reads = string_to_readset(
        """
      1111
         111
            1111
    """,
        source_id=1,
    )

    for read in more_reads:
        readset.add(read)

    selected_reads = readselection(readset, max_cov=2, preferred_source_ids=None, bridging=True)
    assert selected_reads == set([1, 2, 3]), str(selected_reads)

    selected_reads = readselection(readset, max_cov=2, preferred_source_ids=set([3]), bridging=True)
    assert selected_reads == set([0, 1, 3]), str(selected_reads)


# fmt: off
# TODO: the below test case seems to be incomplete
# def test_tuple_scores():
# 'Only example at the moment '
# reads = """
# 1  11010
# 00 00101
# 001 01110
# 1    111
# """
# weights = """
# 2  13112
# 11 23359
# 223 56789
# 2    111
# """
# rs = string_to_readset(reads, weights)
# selected_reads, skipped_reads = readselection(reads, max_cov = 2, bridging= False)
# selected_reads, skipped_reads = readselection(reads, max_cov = 2, bridging= True)

# fmt: on



"""
Remove phasing information from a VCF file

This script removes all types of phasing information from the input VCF and
prints out the modified VCF to standard output. The modifications are:

- The HP, PS and PQ tags are removed
- Phasing in the GT tag (using pipe notation) is removed. The genotypes are
  sorted in ascending order. For example, a GT value of '1|0' is converted
  to '0/1'.

It is not an error if no phasing information was found.
"""
import sys
import logging
from pysam import VariantFile


logger = logging.getLogger(__name__)

TAGS_TO_REMOVE = frozenset(("HP", "PQ", "PS"))


def add_arguments(parser):
    add = parser.add_argument
    add("vcf", metavar="VCF", help='VCF file. Use "-" to read from standard input')


def unphase_header(header):
    for hr in header.records:
        if hr.key == "phasing":
            hr.remove()
            break

    for tag in TAGS_TO_REMOVE:
        if tag in header.formats:
            header.formats.remove_header(tag)


def run_unphase(vcf_path, outfile):
    """
    Read a VCF file, remove phasing information, and write the result to
    outfile, which must be a file-like object.
    """
    if vcf_path == "-":
        reader = VariantFile(sys.stdin)
    else:
        reader = VariantFile(vcf_path)

    unphase_header(reader.header)
    with VariantFile(outfile, mode="w", header=reader.header) as writer:
        for record in reader:
            for tag in TAGS_TO_REMOVE:
                if tag in record.format:
                    del record.format[tag]
            for call in record.samples.values():
                if (
                    call["GT"] is not None
                    and call["GT"][0] is not None
                    and call["GT"][1] is not None
                ):
                    call["GT"] = sorted(call["GT"])
                call.phased = False
            writer.write(record)


def main(args):
    run_unphase(args.vcf, sys.stdout)



import logging
from abc import ABC, abstractmethod
from typing import Dict

from math import log

import networkx as nx

from whatshap.core import Read, ReadSet

logger = logging.getLogger(__name__)


class ReadMergerBase(ABC):
    @abstractmethod
    def merge(self, readset: ReadSet) -> ReadSet:
        pass


class ReadMerger(ReadMergerBase):
    def __init__(
        self, error_rate: float, max_error_rate: float, positive_threshold, negative_threshold
    ):
        """
        error_rate: the probability that a nucleotide is wrong
        max_error_rate: the maximum error rate of any edge of the read
            merging graph allowed before we discard it
        positive_threshold: The threshold of the ratio between the probabilities
            that a pair of reads come from the same haplotype and different
            haplotypes
        negative_threshold: The threshold of the ratio between the probabilities
            that a pair of reads come from the same haplotype and different haplotypes.
        """
        self._error_rate = error_rate
        self._max_error_rate = max_error_rate
        self._positive_threshold = positive_threshold
        self._negative_threshold = negative_threshold

    def merge(self, readset: ReadSet) -> ReadSet:
        """
        Return a set of reads after merging together subsets of reads
        (into super reads) from an input readset according to a
        probabilistic model of how likely sets of reads are to appear
        together on one haplotype and on opposite haplotypes.

        readset -- the input .core.ReadSet object
        """
        logger.info(
            "Merging %d reads with error rate %.2f, maximum error rate %.2f, "
            "positive threshold %d and negative threshold %d ...",
            len(readset),
            self._error_rate,
            self._max_error_rate,
            self._positive_threshold,
            self._negative_threshold,
        )
        logger.debug("Merging started.")
        gblue = nx.Graph()
        gnotblue = nx.Graph()

        # Probability that any nucleotide is wrong
        error_rate = self._error_rate
        logger.debug("Error Rate: %s", error_rate)

        # If an edge has too many errors, we discard it since it is not reliable
        logger.debug("Max Error Rate: %s", self._max_error_rate)

        # Threshold of the ratio between the probabilities that the two reads come from
        # the same side or from different sides
        thr = self._positive_threshold
        logger.debug("Positive Threshold: %s", thr)

        # Threshold_neg is a more conservative threshold for the evidence
        # that two reads should not be clustered together.
        thr_neg = self._negative_threshold
        logger.debug("Negative Threshold: %s", thr_neg)

        thr_diff = 1 + int(log(thr, (1 - error_rate) / (error_rate / 3)))
        thr_neg_diff = 1 + int(log(thr_neg, (1 - error_rate) / (error_rate / 3)))
        logger.debug("Thr. Diff.: %s - Thr. Neg. Diff.: %s", thr_diff, thr_neg_diff)

        logger.debug("Start reading the reads...")
        reads = []
        queue = {}
        for i, read in enumerate(readset):
            alleles = []
            orgn = []
            for variant in read:
                position = variant.position
                allele = variant.allele
                quality = variant.quality

                orgn.append((position, allele, quality))
                assert allele in (0, 1)
                alleles.append(allele)
            reads.append(orgn)

            begin = read[0].position
            end = begin + len(alleles)
            gblue.add_node(i, begin=begin, end=end)
            gnotblue.add_node(i, begin=begin, end=end)
            queue[i] = {"begin": begin, "end": end, "alleles": alleles}
            for x in [id for id in queue.keys() if queue[id]["end"] <= begin]:  # type: ignore
                del queue[x]
            for j in queue.keys():
                if i == j:
                    continue
                match, mismatch = eval_overlap(queue[j], queue[i])
                if (
                    match + mismatch >= thr_neg_diff
                    and min(match, mismatch) / (match + mismatch) <= self._max_error_rate
                    and match - mismatch >= thr_diff
                ):
                    gblue.add_edge(j, i, match=match, mismatch=mismatch)
                    if mismatch - match >= thr_neg_diff:
                        gnotblue.add_edge(j, i, match=match, mismatch=mismatch)

        logger.debug("Finished reading the reads.")
        logger.debug("Number of reads: %s", len(reads))
        logger.debug("Blue Graph")
        logger.debug(
            "Nodes: %s - Edges: %s - ConnComp: %s",
            nx.number_of_nodes(gblue),
            nx.number_of_edges(gblue),
            len(list(nx.connected_components(gblue))),
        )
        logger.debug("Non-Blue Graph")
        logger.debug(
            "Nodes: %s - Edges: %s - ConnComp: %s",
            nx.number_of_nodes(gnotblue),
            nx.number_of_edges(gnotblue),
            len(list(nx.connected_components(gnotblue))),
        )

        # We consider the notblue edges as an evidence that two reads
        # should not be merged together
        # Since we want to merge each blue connected components into
        # a single superread, we check each notblue edge (r1, r2) and
        # we remove some blue edges so that r1 and r2 are not in the
        # same blue connected component

        blue_component = {}
        current_component = 0
        for conncomp in nx.connected_components(gblue):
            for v in conncomp:
                blue_component[v] = current_component
            current_component += 1

        for (u, v) in gnotblue.edges():
            if blue_component[u] != blue_component[v]:
                # Keep only the notblue edges that are inside a blue connected component
                continue
            while v in nx.node_connected_component(gblue, u):
                path = nx.shortest_path(gblue, source=u, target=v)
                # Remove the edge with the smallest support
                # A better strategy is to weight each edge with -log p
                # and remove the minimum (u,v)-cut
                w, x = min(
                    zip(path[:-1], path[1:]),
                    key=lambda p: gblue[p[0]][p[1]]["match"] - gblue[p[0]][p[1]]["mismatch"],
                )
                gblue.remove_edge(w, x)

        # Merge blue components (somehow)
        logger.debug("Started Merging Reads...")
        superreads: Dict = {}  # superreads given by the clusters (if clustering)
        representative = {}  # cluster representative of a read in a cluster

        for cc in nx.connected_components(gblue):
            if len(cc) == 1:
                continue
            r = min(cc)
            superreads[r] = {}
            for i in cc:
                representative[i] = r

        for id in range(len(reads)):
            if id in representative:
                for position, allele, quality in reads[id]:
                    r = representative[id]
                    if position not in superreads[r]:
                        superreads[r][position] = [0, 0]
                    superreads[r][position][allele] += quality

        merged_reads = ReadSet()
        readn = 0
        for id in range(len(reads)):
            read = Read(f"read{readn}")
            readn += 1
            if id in representative:
                if id == representative[id]:
                    for position in sorted(superreads[id]):
                        z = superreads[id][position]
                        allele = 0 if z[0] >= z[1] else 1
                        read.add_variant(position, allele, abs(z[1] - z[0]))
                    merged_reads.add(read)
            else:
                for position, allele, quality in reads[id]:
                    read.add_variant(position, allele, quality)
                merged_reads.add(read)

        logger.debug("Finished merging reads.")
        logger.info(
            "... after merging: merged %d reads into %d reads", len(readset), len(merged_reads)
        )

        return merged_reads


class DoNothingReadMerger(ReadMergerBase):
    def merge(self, readset):
        return readset


def eval_overlap(n1, n2):
    """
    Return a tuple containing the number of matches (resp.,
    mismatches) between a pair (n1,n2) of overlapping reads
    """
    hang1 = n2["begin"] - n1["begin"]
    overlap = zip(n1["alleles"][hang1:], n2["alleles"])
    match = mismatch = 0
    for (c1, c2) in overlap:
        if c1 == c2:
            match += 1
        else:
            mismatch += 1
    return match, mismatch



# flake8: noqa
import itertools as it
from math import ceil
from copy import deepcopy

import logging
from whatshap.core import ReadSet
from whatshap.cli.compare import compute_switch_flips_poly_bt
from whatshap.polyphaseutil import get_coverage

"""
This class is exclusively used for debugging and development.
"""

logger = logging.getLogger(__name__)


def draw_plots(
    block_readsets,
    clustering,
    threading,
    haplotypes,
    cut_positions,
    phasable_variant_table,
    plot_clusters,
    plot_threading,
    output,
):
    # Plot options
    logger.info("Generating plots ...")
    combined_readset = ReadSet()
    for block_readset in block_readsets:
        for read in block_readset:
            combined_readset.add(read)
    if plot_clusters:
        draw_clustering(
            combined_readset,
            clustering,
            phasable_variant_table,
            output + ".clusters.pdf",
            genome_space=False,
        )
    if plot_threading:
        coverage = get_coverage(combined_readset, clustering)
        draw_threading(
            combined_readset,
            clustering,
            coverage,
            threading,
            cut_positions,
            haplotypes,
            phasable_variant_table,
            output + ".threading.pdf",
        )


"""
This method only works for a test dataset, for which the true haplotype of read was encoded
into its name. For any other read name, it just returns -1 for unknown haplotype
"""


def parse_haplotype(name):
    try:
        tokens = name.split("_")
        if tokens[-2] == "HG00514" and tokens[-1] == "HAP1":
            return 0
        elif tokens[-2] == "HG00514" and tokens[-1] == "HAP2":
            return 1
        elif tokens[-2] == "NA19240" and tokens[-1] == "HAP1":
            return 2
        elif tokens[-2] == "NA19240" and tokens[-1] == "HAP2":
            return 3
        elif tokens[-2] == "HG00733" and tokens[-1] == "HAP1":
            return 4
        elif tokens[-2] == "HG00733" and tokens[-1] == "HAP2":
            return 5
    except:
        pass
    return -1


def avg_readlength(readset):
    # Estiamtes the average read length in base pairs
    if len(readset) > 0:
        return sum(read[-1].position - read[0].position for read in readset) / len(readset)
    else:
        return 0


def get_phase(readset, var_table):
    tmp_table = deepcopy(var_table)
    tmp_table.subset_rows_by_position(readset.get_positions())
    try:
        phase_rows = [variant.phase for variant in tmp_table.phases[0]]
    except AttributeError as e:
        return None
    return [[row[i] for row in phase_rows] for i in range(len(phase_rows[0]))]


def draw_clustering(readset, clustering, var_table, path, genome_space=False):
    try:
        import matplotlib

        matplotlib.use("agg")
        import matplotlib.pyplot as plt
        import matplotlib.patches as mpatches
        from pylab import savefig

        # Sort a deep copy of clustering
        clusters = sorted(
            deepcopy(clustering),
            key=lambda x: min(readset[i][0].position for i in x) if len(x) > 0 else 0,
        )

        # Construct real and predicted haplotype per read
        true_hap = [parse_haplotype(read.name) for read in readset]

        # Map variant positions to [0,l)
        index = {}
        rev_index = []
        num_vars = 0
        min_pos = float("inf")
        max_pos = 0
        for position in readset.get_positions():
            index[position] = num_vars
            rev_index.append(position)
            num_vars += 1
            min_pos = min(min_pos, position)
            max_pos = max(max_pos, position)

        min_pos = min(readset.get_positions()) if genome_space else 0
        max_pos = max(readset.get_positions()) if genome_space else num_vars

        # Plot heatmaps
        fig = plt.figure(figsize=(num_vars / 40, len(readset) / 40), dpi=100)
        legend_handles = {}
        y_offset = 0
        y_margin = 5

        # Plot haplotype dissimilarity
        if var_table != None:
            plot_haplotype_dissimilarity(
                legend_handles,
                y_offset,
                y_margin,
                index,
                rev_index,
                readset,
                var_table,
                genome_space,
            )

        y_offset = 0

        # Plot heatmaps
        for c_id in range(0, len(clusters)):
            if len(clusters[c_id]) < 5:
                continue
            read_id = 0
            for read in clusters[c_id]:
                start = index[readset[read][0].position]
                end = index[readset[read][-1].position]
                read_id += 1

                color_code = "C" + str(true_hap[read]) if true_hap[read] != -1 else "black"
                if color_code not in legend_handles:
                    legend_handles[color_code] = mpatches.Patch(
                        color=color_code, label=true_hap[read]
                    )
                if genome_space:
                    plt.hlines(
                        y=read_id + y_offset,
                        xmin=rev_index[start],
                        xmax=rev_index[end],
                        color=color_code,
                    )
                else:
                    plt.hlines(y=read_id + y_offset, xmin=start, xmax=end, color=color_code)

            y_offset += len(clusters[c_id]) + y_margin
            plt.hlines(y=y_offset, xmin=min_pos, xmax=max_pos, color=(0.5, 0.5, 0.5, 0.5))
            y_offset += y_margin

        plt.legend(handles=legend_handles.values(), loc="lower center", ncol=len(legend_handles))
        axes = plt.gca()
        axes.set_xlim([min_pos, max_pos])
        fig.savefig(path)
        fig.clear()

    except ImportError:
        logger.error("Plotting read clusters requires matplotlib to be installed")


def plot_haplotype_dissimilarity(
    legend_handles, y_offset, y_margin, index, rev_index, readset, var_table, genome_space=False
):
    try:
        import matplotlib

        matplotlib.use("agg")
        import matplotlib.pyplot as plt
        import matplotlib.patches as mpatches

        num_vars = len(readset.get_positions())
        min_pos = min(readset.get_positions()) if genome_space else 0
        max_pos = max(readset.get_positions()) if genome_space else num_vars
        readlen = avg_readlength(readset)

        # Plot heatmaps
        y_offset = 0
        y_margin = 5

        # Plot haplotype dissimilarity
        phase_vectors = get_phase(readset, var_table)
        if not phase_vectors:
            # No phasing information available
            return

        chunk = 24
        padding = int(
            readlen // 6
        )  # dissimilarity of position i is averaged over interval of 2*padding base pairs (not positions)

        if genome_space:
            # get variant density
            dens_pos = list(range(min_pos + padding, max_pos - padding, 2 * padding))
            dens_list = [
                len(list(filter(lambda x: pos - padding <= x <= pos + padding, rev_index)))
                for pos in dens_pos
            ]
            max_dens = max(dens_list)
            print("max_dens=" + str(max_dens))

            y_offset -= 104 + y_margin
            plt.hlines(y=y_offset, xmin=min_pos, xmax=max_pos, color="black", lw=1)
            plt.hlines(y=y_offset + 104, xmin=min_pos, xmax=max_pos, color="black", lw=1)
            plt.plot(
                dens_pos, [100 * x / max_dens + y_offset for x in dens_list], lw=1, color="blue"
            )

        # determines for each position, over which interval of positions the average must be taken
        intervals = []
        for i in range(num_vars):
            left = right = i
            pos = rev_index[i]
            while left - 1 >= 0 and rev_index[left - 1] >= pos - padding:
                left -= 1
            while right + 1 < num_vars and rev_index[right + 1] <= pos + padding:
                right += 1
            intervals.append([left, right])

        # One plot for each pair of haplotypes
        for i, j in it.combinations(range(len(phase_vectors)), 2):
            y_offset -= 104 + y_margin
            colors = ["C" + str(i), "C" + str(j)]
            if colors[0] not in legend_handles:
                legend_handles[colors[0]] = mpatches.Patch(color=colors[0], label=i)
            if colors[1] not in legend_handles:
                legend_handles[colors[1]] = mpatches.Patch(color=colors[1], label=j)
            dist = [
                y_offset + 2 + 100 * v
                for v in haplodist(phase_vectors[i], phase_vectors[j], intervals)
            ]
            plt.hlines(y=y_offset, xmin=min_pos, xmax=max_pos, color="black", lw=1)
            plt.hlines(y=y_offset + 104, xmin=min_pos, xmax=max_pos, color="black", lw=1)
            for k in range(ceil(num_vars / chunk)):
                start = k * chunk
                end = min(num_vars, (k + 1) * chunk + 1)
                if genome_space:
                    plt.plot(rev_index[start:end], dist[start:end], lw=1, color=colors[k % 2])
                else:
                    plt.plot(list(range(start, end)), dist[start:end], lw=1, color=colors[k % 2])

    except ImportError:
        logger.error("Plotting haplotype dissimilarities requires matplotlib to be installed")


def haplodist(h1, h2, intervals):
    if len(h1) != len(h2):
        return -1
    n = len(h1)
    return [
        relative_hamming_dist(
            h1[intervals[i][0] : min(n, intervals[i][1] + 1)],
            h2[intervals[i][0] : min(n, intervals[i][1] + 1)],
        )
        for i in range(0, n)
    ]


def relative_hamming_dist(seq1, seq2):
    if len(seq1) != len(seq2):
        return -1
    else:
        return sum(1 for i in range(len(seq1)) if seq1[i] != seq2[i]) / len(seq1)


def draw_threading(
    readset, clustering, coverage, paths, cut_positions, haplotypes, var_table, path
):
    try:
        import matplotlib

        matplotlib.use("agg")
        import matplotlib.pyplot as plt
        import matplotlib.patches as mpatches
        from pylab import savefig

        assert len(paths) > 0
        ploidy = len(paths[0])
        assert ploidy >= 2
        num_c = len(clustering)
        assert num_c > ploidy
        num_vars = len(coverage)

        # Detect relevant clusters
        c_map = {}
        all_threaded = set()
        for p in range(ploidy):
            for pos in range(len(paths)):
                all_threaded.add(paths[pos][p])
        c_list = sorted(all_threaded)
        for i, c_id in enumerate(c_list):
            c_map[c_id] = i
        num_c = len(c_list)

        # Setup figure
        fig = plt.figure(figsize=(num_vars / 40, num_c / 4), dpi=100)
        legend_handles = {}
        x_scale = 1
        y_margin = 0.1
        y_offset = 0
        c_height = 0.9

        # Plot cut positions
        cut_pos = cut_positions + [num_vars]
        for pos in cut_pos:
            plt.vlines(
                x=pos, ymin=0, ymax=num_c * (c_height + y_margin), color="lightgray", alpha=0.3
            )

        # Plot cluster coverage
        xs = list(range(num_vars))
        for c_id in range(num_c):
            min_pos = num_vars
            max_pos = 0
            for pos in range(num_vars):
                if c_list[c_id] in coverage[pos] and coverage[pos][c_list[c_id]] > 0:
                    min_pos = min(min_pos, pos)
                    max_pos = max(max_pos, pos)
            ys = [
                y_offset + c_height * coverage[pos][c_list[c_id]]
                if c_list[c_id] in coverage[pos]
                else y_offset
                for pos in range(min_pos, max_pos + 1)
            ]
            plt.fill_between(x=xs[min_pos : max_pos + 1], y1=ys, y2=y_offset, color="gray")
            plt.hlines(
                y=y_offset + c_height + y_margin / 2,
                xmin=0,
                xmax=x_scale * num_vars - 1,
                color="lightgray",
                alpha=0.5,
            )
            y_offset += c_height + y_margin

        # Plot paths
        for p in range(ploidy):
            legend_handles["C" + str(p)] = mpatches.Patch(
                color="C" + str(p), label=("Hap " + str(p))
            )
            current = paths[0][p]
            start = 0
            for pos in range(1, len(paths)):
                if paths[pos][p] != current:
                    plt.hlines(
                        y=(c_map[current] + 0.25 + p / ploidy * 0.5) * (c_height + y_margin),
                        xmin=x_scale * start,
                        xmax=x_scale * pos,
                        color="C" + str(p),
                        alpha=0.9,
                    )
                    current = paths[pos][p]
                    start = pos
            plt.hlines(
                y=(c_map[current] + 0.25 + p / ploidy * 0.5) * (c_height + y_margin),
                xmin=x_scale * start,
                xmax=x_scale * num_vars - 1,
                color="C" + str(p),
                alpha=0.9,
            )

        # Plot switch flip errors
        # print(cut_positions)
        # print(cut_pos)
        # print(haplotypes)

        # If we have ground truth, retrieve it
        compare = True
        try:
            phase_vectors = get_phase(readset, var_table)
            truth = []
            assert len(phase_vectors) == ploidy
            for k in range(ploidy):
                truth.append("".join(map(str, phase_vectors[k])))
        except:
            compare = False

        if compare:
            for i in range(len(cut_pos) - 1):
                block1 = [h[cut_pos[i] : min(len(paths), cut_pos[i + 1])] for h in truth]
                block2 = [h[cut_pos[i] : min(len(paths), cut_pos[i + 1])] for h in haplotypes]

                (
                    switchflips,
                    switches_in_column,
                    flips_in_column,
                    poswise_config,
                ) = compute_switch_flips_poly_bt(
                    block1,
                    block2,
                    report_error_positions=True,
                    switch_cost=1 + 1 / (num_vars * ploidy),
                )
                for pos, e in enumerate(switches_in_column):
                    if e > 0:
                        plt.vlines(
                            x=cut_pos[i] + pos,
                            ymax=-y_margin,
                            ymin=-y_margin - c_height * e,
                            color="blue",
                            alpha=0.6,
                        )
                        switches = [
                            j
                            for j in range(ploidy)
                            if poswise_config[pos][j] != poswise_config[pos - 1][j]
                        ]
                        for h in switches:
                            c_id = c_map[paths[cut_pos[i] + pos][h]]
                            plt.vlines(
                                x=cut_pos[i] + pos,
                                ymax=c_id + 0.95 * c_height,
                                ymin=c_id + 0.05 * c_height,
                                color="black",
                                alpha=0.3,
                            )
                for pos, flipped in enumerate(flips_in_column):
                    if len(flipped) == 0:
                        continue
                    if cut_pos[i] + pos >= len(paths):
                        continue
                    plt.vlines(
                        x=cut_pos[i] + pos,
                        ymax=-y_margin,
                        ymin=-y_margin - c_height * len(flipped),
                        color="orange",
                        alpha=0.6,
                    )
                    for h in flipped:
                        c_id = c_map[paths[cut_pos[i] + pos][h]]
                        plt.hlines(
                            y=(c_id + 0.25 + h / ploidy * 0.5) * (c_height + y_margin),
                            xmin=cut_pos[i] + pos - 0.5,
                            xmax=cut_pos[i] + pos + 0.5,
                            color="black",
                            alpha=0.6,
                        )

        # plt.legend(handles=legend_handles.values(), loc='lower center', ncol=len(legend_handles))
        axes = plt.gca()
        axes.set_xlim([0, num_vars - 1])
        fig.savefig(path)
        fig.clear()

    except ImportError:
        logger.error("Plotting haplotype threading requires matplotlib to be installed")



class CovMonitor:
    # TODO: This is a most simple, naive implementation. Could do this smarter.
    # TODO : Maybe change the coverage monitor from the usage of vcf_indices to the real varaint positions

    def __init__(self, length):
        self.coverage = [0] * length

    def max_coverage_in_range(self, begin, end):
        return max(self.coverage[begin:end])

    def add_read(self, begin, end):
        for i in range(begin, end):
            self.coverage[i] += 1



# cython: language_level=3

from libcpp cimport bool
from libc.stdint cimport uint32_t, uint64_t
from . cimport cpp


cdef class NumericSampleIds:
	cdef dict mapping
	cdef bool frozen


cdef class Read:
	cdef cpp.Read *thisptr
	cdef bool ownsptr


cdef class ReadSet:
	cdef cpp.ReadSet *thisptr


cdef class Pedigree:
	cdef cpp.Pedigree *thisptr
	cdef NumericSampleIds numeric_sample_ids


cdef class PedigreeDPTable:
	cdef cpp.PedigreeDPTable *thisptr
	cdef Pedigree pedigree


cdef class PhredGenotypeLikelihoods:
	cdef cpp.PhredGenotypeLikelihoods *thisptr
	
	
cdef class Genotype:
	cdef cpp.Genotype *thisptr
	cdef uint64_t index
	cdef uint32_t ploidy


cdef class GenotypeDPTable:
	cdef cpp.GenotypeDPTable *thisptr
	cdef Pedigree pedigree
	cdef NumericSampleIds numeric_sample_ids


cdef class HapChatCore:
	cdef cpp.HapChatCore *thisptr



"""
Convert hapCUT output format to VCF

HapCUT’s output is combined with the original VCF and
then written as phased VCF to standard output.

HapCUT 1 and 2 are supported.

HapCUT’s output file format is explained at
<https://github.com/vibansal/hapcut#format-of-input-and-output-files>

HapCUT2’s output format is documented at
<https://github.com/pjedge/hapcut2#output-format>
"""
import logging
import re
import sys
from collections import namedtuple
import itertools
from contextlib import ExitStack

from whatshap.cli import CommandLineError
from whatshap.vcf import PhasedVcfWriter
from whatshap.core import Read
from whatshap import __version__

logger = logging.getLogger(__name__)


def add_arguments(parser):
    add = parser.add_argument
    add(
        "-o",
        "--output",
        default=sys.stdout,
        help="Output VCF file. If omitted, use standard output.",
    )
    add("vcf", metavar="VCF", help="VCF file")
    add("hapcut", metavar="HAPCUT-RESULT", help="hapCUT result file")


HapCutVariant = namedtuple(
    "HapCutVariant", ["chromosome", "position", "haplotype1", "haplotype2", "component_id"]
)


class ParseError(Exception):
    pass


class HapCutParser:
    """Parse HapCUT results"""

    # Example for HapCUT
    #
    # BLOCK: offset: 282 len: 9 phased: 7 SPAN: 5896 MECscore 0.00 fragments 7
    # 282    0    1    1    1065296    T    C    1|0    1,0:-0.0,-0.0,-1.3:0.0:0.0
    # 283    0    1    1    1066259    G    C    1|0    2,0:-0.1,-0.1,-2.7:0.0:0.0
    # 284    0    1    1    1066282    A    G    1|0    3,0:-0.1,-0.2,-4.2:0.0:0.0
    # ********
    # BLOCK: offset: 291 len: 3 phased: 3 SPAN: 7788 MECscore 0.00 fragments 3
    # 291    1    0    1    1072498    G    C    1|0    0,1:-1.4,-0.0,-0.0:0.0:0.0
    # 292    1    0    1    1077064    C    A    1|0    0,2:-2.6,-0.1,-0.0:0.0:0.0
    # 293    1    0    1    1080286    G    A    1|0    0,3:-4.1,-0.1,-0.1:0.1:0.0
    #
    # HapCUT2 output format is slightly different:
    #
    # BLOCK: offset: 306 len: 37 phased: 27 SPAN: 38662 fragments 140
    # 306  0  1  1  1065296 T  C  0/1  0  0.000000  -0.000000
    # 307  0  1  1  1066259 G  C  0/1  0  0.000000  0.000000

    block_re = re.compile(
        r"BLOCK: "
        r"offset: (?P<offset>\d+) "
        r"len: (?P<len>\d+) "
        r"phased: (?P<phased>\d+) "
        r"SPAN: (?P<span>\d+) "
        r"(MECscore (?P<mecscore>\d+\.\d+) )?fragments (?P<fragments>\d+)"
    )

    def __init__(self, file):
        self._file = file

    def __iter__(self):
        """Yield (chromosome, blocks) pairs"""
        yield from self._by_chromosome()

    def parse_blocks(self):
        """
        Yield a list of HapCutVariant objects for each connected component ('block')
        """
        state = "BLOCK"  # DFA states are BLOCK and VARIANT; they describe what we expect next
        block = []
        for line in self._file:
            if state == "BLOCK":
                state = "VARIANT"
                if not line.startswith("BLOCK:"):
                    raise ParseError('Expected a new block (line starting with "BLOCK:")')
                m = self.block_re.match(line)
                if not m:
                    raise ParseError("BLOCK line malformed")
            elif state == "VARIANT":
                if line.startswith("********"):
                    # End marker reached, yield the list of variants in this connected component
                    if block:
                        yield block
                    state = "BLOCK"
                    block = []
                else:
                    fields = line.strip().split()
                    if len(fields) not in (9, 11):
                        raise ParseError(
                            "Expected nine fields (for hapCUT 1) "
                            "or eleven fields (for hapCUT 2) in variant line"
                        )
                    (
                        variant_id,
                        haplotype_1,
                        haplotype_2,
                        chromosome,
                        position,
                        reference_allele,
                        alternative_allele,
                        genotype,
                    ) = fields[:8]

                    if len(fields) == 9:  # hapCUT 1
                        # The last fields are not actually used, we just check
                        # whether they are formatted correctly
                        rest = fields[8]
                        fields = rest.split(":")
                        if len(fields) == 5:
                            if not fields[-1] == "FV":
                                raise ParseError('Expected "FV" after last colon')
                            fields = fields[:-1]
                        if not len(fields) == 4:
                            raise ParseError("Too few elements in last (colon-separated) field")
                        # allele_counts, genotype_likelihoods, delta, mec_variant = fields
                        # allele_counts = [ int(s) for s in allele_counts.split(',') ]
                        # genotype_likelihoods = [ float(s) for s in genotype_likelihoods.split(',') ]
                        # delta = float(delta)
                        # mec_variant = float(mec_variant)
                    elif len(fields) == 11:
                        # pruned, switch_qual, flip_qual = fields[8:]
                        pass
                    if haplotype_1 == "-" or haplotype_2 == "-":
                        # This happens in hapCUT 2 sometimes
                        continue
                    _ = int(variant_id)
                    haplotype_1 = int(haplotype_1)
                    haplotype_2 = int(haplotype_2)
                    position = int(position) - 1
                    component_id = block[0].position if block else position
                    variant = HapCutVariant(
                        chromosome, position, haplotype_1, haplotype_2, component_id
                    )
                    block.append(variant)
        if len(block) > 0:
            yield block

    def _by_chromosome(self):
        for chromosome, block in itertools.groupby(self.parse_blocks(), lambda b: b[0].chromosome):
            yield chromosome, list(block)


def run_hapcut2vcf(hapcut, vcf, output=sys.stdout):
    command_line = "(whatshap {}) {}".format(__version__, " ".join(sys.argv[1:]))
    with ExitStack() as stack:
        if isinstance(output, str):
            output = stack.enter_context(open(output, "w"))

        writer = PhasedVcfWriter(vcf, command_line, out_file=output)
        if len(writer.samples) > 1:
            # This would be easy to support with a --sample command-line parameter,
            # but hapCUT does not seem to support multi-sample VCFs, so something
            # must be wrong anyway.
            raise CommandLineError("There is more than one sample in this VCF")
        sample = writer.samples[0]

        f = stack.enter_context(open(hapcut))
        parser = HapCutParser(f)
        for chromosome, blocks in parser:
            logger.info("Read %d phased blocks for chromosome %s", len(blocks), chromosome)

            # Build one read for each haplotype and the connected components
            haplotypes = [Read(str(i)) for i in (1, 2)]
            components = dict()
            for block in blocks:
                for variant in block:
                    haplotypes[0].add_variant(variant.position, variant.haplotype1, 0)
                    haplotypes[1].add_variant(variant.position, variant.haplotype2, 0)
                    components[variant.position] = variant.component_id

            sample_superreads = {sample: haplotypes}
            sample_components = {sample: components}
            writer.write(chromosome, sample_superreads, sample_components)


def main(args):
    run_hapcut2vcf(**vars(args))



from typing import Iterable, Tuple, Sequence

from pysam import AlignedSegment

from .vcf import VcfVariant

def _iterate_cigar(
    variants: Sequence[VcfVariant],
    j: int,
    bam_read: AlignedSegment,
    cigartuples: Iterable[Tuple[int, int]],
) -> Iterable[Tuple[int, int, int, int]]: ...



from whatshap.merge import ReadMerger
from whatshap.testhelpers import string_to_readset


def assert_variants(reads, expected):
    # assert that the lists of variants (pos, allele, weight) are identical
    for read, expected_read in zip(reads, expected):
        assert list(read) == list(expected_read)


def test_read_merging():
    reads = string_to_readset(
        """
      0 000000
      111
      11 00111101
      0 00000
    """,
        """
      1 523428
      714
      86 03158958
      8 46626
    """,
    )

    merger = ReadMerger(0.15, 0.25, 100000, 1000)
    merged_reads = merger.merge(reads)
    # default parameter settings

    expected = string_to_readset(
        """
      0 000000
      111
      11 00111101
    """,
        """
      9 989688
      714
      86 03158958
    """,
    )

    assert_variants(merged_reads, expected)


def test_read_merging2():
    reads = string_to_readset(
        """
      0 000000
      111
      11 00111101
      0 00000
    """,
        """
      1 523428
      714
      86 03158958
      8 46626
    """,
    )
    merger = ReadMerger(0.5, 0.5, 1000, 100000)
    merged_reads = merger.merge(reads)
    # error rates and thresholds so high that no merging occurs

    assert_variants(merged_reads, reads)



import os
from setuptools import setup, Extension
from distutils.sysconfig import customize_compiler
import Cython.Build
from Cython.Build import cythonize


def CppExtension(name, sources):
    return Extension(
        name,
        sources=sources,
        language="c++",
        extra_compile_args=["-std=c++11", "-Werror=return-type", "-Werror=narrowing"],
        undef_macros=["NDEBUG"],
    )


extensions = [
    CppExtension(
        "whatshap.core",
        sources=[
            "whatshap/core.pyx",
            "src/pedigree.cpp",
            "src/pedigreedptable.cpp",
            "src/pedigreecolumncostcomputer.cpp",
            "src/columnindexingiterator.cpp",
            "src/columnindexingscheme.cpp",
            "src/entry.cpp",
            "src/graycodes.cpp",
            "src/read.cpp",
            "src/readset.cpp",
            "src/columniterator.cpp",
            "src/indexset.cpp",
            "src/genotype.cpp",
            "src/binomial.cpp",
            "src/pedigreepartitions.cpp",
            "src/phredgenotypelikelihoods.cpp",
            "src/genotyper.cpp",
            "src/genotypedistribution.cpp",
            "src/genotypedptable.cpp",
            "src/genotypecolumncostcomputer.cpp",
            "src/backwardcolumniterator.cpp",
            "src/transitionprobabilitycomputer.cpp",
            "src/hapchat/basictypes.cpp",
            "src/hapchat/balancedcombinations.cpp",
            "src/hapchat/binomialcoefficient.cpp",
            "src/hapchat/hapchatcore.cpp",
            "src/hapchat/hapchatcolumniterator.cpp",
        ],
    ),
    CppExtension(
        "whatshap.polyphase_solver",
        sources=[
            "whatshap/polyphase_solver.pyx",
            "src/polyphase/clustereditingsolution.cpp",
            "src/polyphase/clustereditingsolver.cpp",
            "src/polyphase/edgeheap.cpp",
            "src/polyphase/inducedcostheuristic.cpp",
            "src/polyphase/staticsparsegraph.cpp",
            "src/polyphase/switchflipcalculator.cpp",
            "src/polyphase/trianglesparsematrix.cpp",
            "src/polyphase/readscoring.cpp",
            "src/polyphase/haplothreader.cpp",
            "src/polyphase/tupleconverter.cpp",
        ],
    ),
    CppExtension("whatshap.readselect", sources=["whatshap/readselect.pyx"]),
    CppExtension("whatshap.priorityqueue", sources=["whatshap/priorityqueue.pyx"]),
    CppExtension("whatshap.align", sources=["whatshap/align.pyx"]),
    CppExtension("whatshap._variants", sources=["whatshap/_variants.pyx"]),
]


class BuildExt(Cython.Build.build_ext):
    def build_extensions(self):
        # Remove the warning about “-Wstrict-prototypes” not being valid for C++,
        # see http://stackoverflow.com/a/36293331/715090
        customize_compiler(self.compiler)
        if self.compiler.compiler_so[0].endswith("clang"):
            # Clang needs this option in order to find the unordered_set header
            print("detected clang, using option -stdlib=libc++")
            self.compiler.compiler_so.append("-stdlib=libc++")
        try:
            self.compiler.compiler_so.remove("-Wstrict-prototypes")
        except (AttributeError, ValueError):
            pass
        super().build_extensions()


# Avoid compilation if we are being installed within Read The Docs
if os.environ.get("READTHEDOCS") == "True":
    cmdclass = {}
    ext_modules = []
    install_requires = []
else:
    cmdclass = {"build_ext": BuildExt}
    ext_modules = extensions
    install_requires = [
        "pysam>=0.18.0",
        "pyfaidx>=0.5.5.2",
        "networkx",
        "biopython>=1.73",  # pyfaidx needs this for reading bgzipped FASTA files
        "scipy",
        "xopen>=1.2.0",
    ]

setup(
    use_scm_version={"write_to": "whatshap/_version.py"},
    cmdclass=cmdclass,
    ext_modules=cythonize(ext_modules),
    install_requires=install_requires,
)



"""
Test Read and ReadSet classes
"""
from pytest import raises
from whatshap.core import Read, ReadSet, Variant


def test_read():
    r = Read("name", 15)
    assert r.name == "name"
    assert r.mapqs[0] == 15

    assert r.is_sorted()

    r.add_variant(100, 1, 37)
    r.add_variant(23, 0, 99)
    assert not r.is_sorted()
    r.sort()
    assert r.is_sorted()

    assert 100 in r
    assert 23 in r
    assert 22 not in r
    assert 24 not in r
    assert 1000 not in r
    assert -1000 not in r


def test_read_iteration():
    r = Read("name", 15)
    r.add_variant(100, 1, 37)
    r.add_variant(23, 0, 99)
    v1 = Variant(position=100, allele=1, quality=37)
    v2 = Variant(position=23, allele=0, quality=99)
    variants = list(r)
    assert variants == [v1, v2]
    # negative indices
    assert r[-1] == v2
    assert r[-2] == v1


def test_read_indexerror1():
    r = Read("name", 15)
    r.add_variant(100, 1, 37)
    r.add_variant(23, 0, 99)
    with raises(IndexError):
        _ = r[2]


def test_read_indexerror2():
    r = Read("name", 15)
    r.add_variant(100, 1, 37)
    r.add_variant(23, 0, 99)
    with raises(IndexError):
        _ = r[-3]


def test_empty_readset():
    rs = ReadSet()
    assert len(rs) == 0


def test_readset():
    rs = ReadSet()
    r = Read("Read A", 56)
    r.add_variant(100, 1, 37)
    r.add_variant(101, 0, 18)
    rs.add(r)

    r = Read("Read B", 0)
    r.add_variant(101, 0, 23)
    rs.add(r)

    r = Read("Read C", 17)
    r.add_variant(99, 1, 27)
    r.add_variant(80, 1, 17)
    r[1] = Variant(position=105, allele=0, quality=14)
    rs.add(r)

    assert rs[0].name == "Read A"
    assert rs[1].name == "Read B"
    assert rs[2].name == "Read C"

    rs.sort()

    # should be sorted after finalization
    assert rs[0].name == "Read C"
    assert rs[1].name == "Read A"
    assert rs[2].name == "Read B"

    assert len(rs) == 3

    assert rs.get_positions() == [99, 100, 101, 105]

    r = rs[(0, "Read A")]
    assert r.name == "Read A"
    assert r.mapqs == (56,), str(r.mapqs)

    r = rs[(0, "Read B")]
    assert r.name == "Read B"
    assert r.mapqs == (0,)

    r = rs[(0, "Read C")]
    assert r.name == "Read C"
    assert r.mapqs == (17,)
    assert len(r) == 2
    assert r[0] == Variant(position=99, allele=1, quality=27)
    assert r[1] == Variant(position=105, allele=0, quality=14)


def test_readset2():
    rs = ReadSet()
    rs.add(Read("Read A", 1, 23))
    rs.add(Read("Read A", 2, 70))
    rs.add(Read("Read B", 3, 23))
    assert rs[(23, "Read A")].mapqs == (1,)
    assert rs[(70, "Read A")].mapqs == (2,)
    assert rs[(23, "Read B")].mapqs == (3,)


def test_non_existing_read_name():
    rs = ReadSet()
    r = Read("Read A", 56)
    r.add_variant(100, 1, 37)
    r.add_variant(101, 0, 18)
    rs.add(r)
    with raises(KeyError):
        _ = rs[(0, "foo")]


def test_non_existing_read_name2():
    rs = ReadSet()
    r = Read("Read A", 56, 1)
    r.add_variant(100, 1, 37)
    r.add_variant(101, 0, 18)
    rs.add(r)
    with raises(KeyError):
        _ = rs[(2, "Read A")]


# TODO: Test subset method



from whatshap.cli.find_snv_candidates import run_find_snv_candidates


def test_call(tmpdir):
    output = str(tmpdir.join("output.vcf"))
    run_find_snv_candidates(
        "tests/data/pacbio/reference.fasta",
        "tests/data/pacbio/pacbio.bam",
        datatype="pacbio",
        output=output,
    )
    computed_lines = []
    expected_lines = []
    for line in open(output, "r"):
        if line.startswith("#"):
            continue
        computed_lines.append(line)
    for line in open("tests/data/expected-calls.vcf"):
        if line.startswith("#"):
            continue
        expected_lines.append(line)
    assert computed_lines == expected_lines



import logging

from collections import defaultdict
from .core import Read, ReadSet
from queue import Queue

logger = logging.getLogger(__name__)


def get_position_map(readset):
    """
    Returns a mapping of genome (bp) positions to virtual positions (from 0 to l).
    """
    # Map genome positions to [0,l)
    index = {}
    rev_index = []
    num_vars = 0

    for position in readset.get_positions():
        index[position] = num_vars
        rev_index.append(position)
        num_vars += 1

    return index, rev_index


def get_coverage(readset, clustering):
    """
    Returns a list, which for every position contains a dictionary, mapping a cluster id to
    a relative coverage on this position.
    """
    pos_index, rev_index = get_position_map(readset)
    num_vars = len(pos_index)
    num_clusters = len(clustering)
    coverage = [defaultdict(int) for pos in range(num_vars)]
    coverage_sum = [0 for pos in range(num_vars)]
    for c_id in range(num_clusters):
        for read in clustering[c_id]:
            for pos in [pos_index[var.position] for var in readset[read]]:
                if c_id not in coverage[pos]:
                    coverage[pos][c_id] = 0
                coverage[pos][c_id] += 1
                coverage_sum[pos] += 1

    for pos in range(num_vars):
        for c_id in coverage[pos]:
            coverage[pos][c_id] = coverage[pos][c_id] / coverage_sum[pos]

    return coverage


def haplotypes_to_strings(haplotypes):
    haplostrings = []
    for i, h in enumerate(haplotypes):
        haplostrings.append("".join(map(lambda x: str(x) if x >= 0 else "n", h)))
    return haplostrings


def split_readset(readset, ext_block_starts):
    """
    Creates one sub-readset for every block. Reads which cross block borders are also split, so
    parts of them appear in multiple blocks. Reads inside a sub-readset are trimmed, such that they
    do not contain variants outside of their associated blocks.
    """

    index, rev_index = get_position_map(readset)
    var_to_block = [0 for _ in range(ext_block_starts[-1])]
    for i in range(len(ext_block_starts) - 1):
        for var in range(ext_block_starts[i], ext_block_starts[i + 1]):
            var_to_block[var] = i

    block_readsets = [ReadSet() for i in range(len(ext_block_starts) - 1)]
    for i, read in enumerate(readset):
        if not read.is_sorted():
            read.sort()
        start = var_to_block[index[read[0].position]]
        end = var_to_block[index[read[-1].position]]
        if start == end:
            # if read lies entirely in one block, copy it into according readset
            block_readsets[start].add(read)
        else:
            # split read by creating one new read for each covered block
            current_block = start
            read_slice = Read(
                name=read.name,
                source_id=read.source_id,
                sample_id=read.sample_id,
                reference_start=read.sample_id,
                BX_tag=read.BX_tag,
            )
            for variant in read:
                if var_to_block[index[variant.position]] != current_block:
                    block_readsets[current_block].add(read_slice)
                    current_block = var_to_block[index[variant.position]]
                    read_slice = Read(
                        name=str(current_block) + "_" + read.name,
                        source_id=read.source_id,
                        sample_id=read.sample_id,
                        reference_start=read.sample_id,
                        BX_tag=read.BX_tag,
                    )
                read_slice.add_variant(variant.position, variant.allele, variant.quality)
            block_readsets[current_block].add(read_slice)
    return block_readsets


def compute_block_starts(readset, ploidy, single_linkage=False):
    """
    Based on the connectivity of the reads, we want to divide the phasing input, as non- or poorly
    connected regions can be phased independently. This is done based on how pairs of variants are
    connected. There are two modes how to decide whether two variants are connected:

    single_linkage=True -- If there exists a read in the readset, which covers both variants, they
                           are connected
    single_linkage=False -- In order to connect two variants, we need at least reads from ploidy-1
                            different haplotypes. Two variants count as connected, if there
                            sufficiently many reads covering both variants, with "sufficient"
                            meaning, that the connecting reads have a chance of at least 98% that
                            they cover at least ploidy-1 haplotypes.

    First, only consecutive pairs are inspected. Then, this connectivity is made transitive, i.e.
    if the pair (A,C) is connected, as well as the pair (B,C), then (A,B) is also connected. If the
    special case occurs, that for three variants (in this order) A and C are connected, but neither
    is connected to variant B in between them, then the variants are still divided as A|B|C, even
    though A and C are actually connected. This is because the following steps require the splits
    to be intervals with no "holes" inside them.
    """

    pos_index, rev_index = get_position_map(readset)
    num_vars = len(pos_index)

    # special case
    if num_vars == 0:
        return []

    # cut threshold
    if ploidy == 2 or single_linkage:
        cut_threshold = 1
    else:
        cut_threshold = ploidy * ploidy
        for i in range(ploidy - 1, ploidy * ploidy):
            # chance to cover at most ploidy-2 haplotypes with i reads
            cut_threshold = i
            if ploidy * pow((ploidy - 2) / ploidy, i) < 0.02:
                cut_threshold = i
                break
    logger.debug(f"Cut position threshold: coverage >= {cut_threshold}")

    # start by looking at neighbouring
    link_to_next = [0 for i in range(num_vars)]
    starts = []
    ends = []
    for read in readset:
        pos_list = [pos_index[var.position] for var in read]
        starts.append(pos_list[0])
        ends.append(pos_list[-1])
        for i in range(len(pos_list) - 1):
            if pos_list[i] + 1 == pos_list[i + 1]:
                link_to_next[pos_list[i]] += 1

    pos_clust = [0 for i in range(num_vars)]
    for i in range(1, num_vars):
        if link_to_next[i - 1] >= cut_threshold:
            pos_clust[i] = pos_clust[i - 1]
        else:
            pos_clust[i] = pos_clust[i - 1] + 1
    num_clust = pos_clust[-1] + 1

    # find linkage between clusters
    link_coverage = [dict() for i in range(num_clust)]
    for i, (start, end) in enumerate(zip(starts, ends)):
        covered_pos_clusts = {pos_clust[pos_index[var.position]] for var in readset[i]}
        for p1 in covered_pos_clusts:
            for p2 in covered_pos_clusts:
                if p2 not in link_coverage[p1]:
                    link_coverage[p1][p2] = 0
                link_coverage[p1][p2] += 1

    # merge clusters
    merged_clust = [-1 for i in range(num_clust)]
    new_num_clust = 0
    for i in range(num_clust):
        if merged_clust[i] >= 0:
            continue
        q = Queue()
        q.put(i)
        merged_clust[i] = new_num_clust
        while not q.empty():
            cur = q.get()
            for linked in link_coverage[cur]:
                if merged_clust[linked] < 0 and link_coverage[cur][linked] >= cut_threshold:
                    q.put(linked)
                    merged_clust[linked] = new_num_clust
        new_num_clust += 1

    # determine cut positions
    cuts = [0]
    for i in range(1, num_vars):
        if merged_clust[pos_clust[i]] != merged_clust[pos_clust[i - 1]]:
            cuts.append(i)

    return cuts


def create_genotype_list(phasable_variant_table, sample):
    """
    Creates a list, which stores a dictionary for every position. The dictionary maps every allele
    to its frequency in the genotype of the respective position.
    """
    all_genotypes = phasable_variant_table.genotypes_of(sample)
    genotype_list = []
    for pos in range(len(all_genotypes)):
        allele_count = dict()
        for allele in all_genotypes[pos].as_vector():
            if allele not in allele_count:
                allele_count[allele] = 0
            allele_count[allele] += 1
        genotype_list.append(allele_count)
    return genotype_list



import os
from abc import ABC
from urllib.parse import urlparse
from typing import Optional, Iterable

import pysam
import logging
import heapq
from collections import defaultdict
from dataclasses import dataclass

logger = logging.getLogger(__name__)


@dataclass
class AlignmentWithSourceID:
    source_id: int
    bam_alignment: pysam.AlignedSegment


class AlignmentFileNotIndexedError(Exception):
    pass


class SampleNotFoundError(Exception):
    pass


class ReferenceNotFoundError(Exception):
    pass


class EmptyAlignmentFileError(Exception):
    pass


def is_local(path):
    return urlparse(path).scheme == ""


class BamReader(ABC):
    pass


class SampleBamReader(BamReader):
    """
    A wrapper for Samfile that provides only those reads from a BAM or CRAM file
    that belong to a specified sample. The BAM/CRAM file must have an index
    (bai/crai).
    """

    def __init__(self, path: str, *, source_id: int = 0, reference: Optional[str] = None):
        """
        path -- URL or path to BAM or CRAM file
        reference -- optional path to FASTA reference for CRAM
        """
        self.source_id: int = source_id
        if reference:
            reference = os.path.abspath(reference)

        self._samfile = pysam.AlignmentFile(path, reference_filename=reference)
        try:
            # TODO
            # multiple_iterators should not be necessary -
            # perhaps a bug in pysam 0.13
            fetcher = self._samfile.fetch(multiple_iterators=True)
        except ValueError:
            raise AlignmentFileNotIndexedError(path)

        # For CRAM files, AlignmentFile.fetch() does not raise an error if the
        # reference could not be retrieved; it just returns an empty list.
        # To detect this situation, we treat all empty alignment files as an error.
        try:
            next(fetcher)
        except StopIteration:
            raise EmptyAlignmentFileError(path) from None
        self._references = frozenset(self._samfile.references)
        self._initialize_sample_to_group_ids()

    def has_reference(self, name: str) -> bool:
        return name in self._references

    def _initialize_sample_to_group_ids(self) -> None:
        """
        Create a dictionary that maps a sample name to a set of read group ids.
        """
        read_groups = self._samfile.header.to_dict().get("RG", [])  # a list of dicts
        logger.debug("Read groups in CRAM/BAM header: %s", read_groups)
        samples = defaultdict(list)
        for read_group in read_groups:
            if "SM" in read_group:
                samples[read_group["SM"]].append(read_group["ID"])
            else:
                logger.warning(
                    'Read group "%s" does not contain an SM field to assign it to a sample.'
                    " Use --ignore-read-groups to use these alignments anyway.",
                    read_group["ID"],
                )
        self._sample_to_group_ids = {id: frozenset(values) for id, values in samples.items()}

    def has_sample(self, sample: str) -> bool:
        """Return whether this file contains reads for the given sample"""
        return sample in self._sample_to_group_ids

    def fetch(self, reference: str, sample: str, start: int = 0, end: Optional[int] = None):
        """
        Yield instances of AlignmentWithSourceID, with source_id value given
        at construction time.
        Raise KeyError if sample not found among samples named in RG header.
        Raise ReferenceNotFoundError when reference is not in the BAM.
        """
        if reference not in self._references:
            # fetch() (below) raises a ValueError when the reference is invalid,
            # but also on other errors.
            raise ReferenceNotFoundError(reference)
        if sample is None:
            for bam_read in self._samfile.fetch(reference, start=start, stop=end):
                yield AlignmentWithSourceID(self.source_id, bam_read)
        else:
            try:
                read_groups = self._sample_to_group_ids[sample]
            except KeyError:
                raise SampleNotFoundError()
            # TODO
            # The multiple_iterators shouldn’t be necessary, but CRAM files
            # don’t work without it in pysam 0.13
            for bam_read in self._samfile.fetch(
                reference, multiple_iterators=True, start=start, stop=end
            ):
                if bam_read.get_tag("RG") in read_groups:
                    yield AlignmentWithSourceID(self.source_id, bam_read)

    def close(self) -> None:
        self._samfile.close()


class ComparableAlignedSegment:
    """
    Heapsort wants to be able to use the less than operator. Native
    AlignedSegment instances do not support this.
    """

    def __init__(self, aligned_segment: pysam.AlignedSegment, source_id: int):
        self.segment: pysam.AlignedSegment = aligned_segment
        self.source_id: int = source_id

    def __lt__(self, other) -> bool:
        self_pos = self.segment.reference_start
        other_pos = other.segment.reference_start
        return (self_pos < other_pos) or (
            self_pos == other_pos and self.source_id < other.source_id
        )


class MultiBamReader(BamReader):
    """
    Read multiple sorted CRAM/BAM files and merge them on the fly.

    To avoid needing to handle renaming of duplicate read groups, this class
    just allows to specify a desired sample name. Doing that filtering here
    is much easier.
    """

    def __init__(self, paths: Iterable[str], *, reference: Optional[str] = None):
        self._readers = []
        for source_id, path in enumerate(paths):
            self._readers.append(SampleBamReader(path, source_id=source_id, reference=reference))

    def fetch(
        self,
        reference: Optional[str] = None,
        sample: Optional[str] = None,
        start: int = 0,
        end: Optional[int] = None,
    ):
        """
        Yield reads from the specified region in all the opened CRAM/BAM files,
        merging them on the fly. Each CRAM/BAM file must be indexed.

        Yields instances of AlignmentWithSourceID, where source_id corresponds to
        index of BAM file name given at construction time.

        If a sample name is given, only reads that belong to that sample are
        returned (the RG tags of each read and the RG header are used for that).
        """
        assert reference is not None

        def make_comparable(reader):
            for alignment in reader.fetch(reference, sample, start, end):
                yield ComparableAlignedSegment(alignment.bam_alignment, alignment.source_id)

        iterators = []
        for reader in self._readers:
            if sample is None or reader.has_sample(sample):
                iterators.append(make_comparable(reader))
        if not iterators:
            raise SampleNotFoundError("Sample not found in any input CRAM/BAM file")
        for it in heapq.merge(*iterators):
            yield AlignmentWithSourceID(it.source_id, it.segment)

    def has_reference(self, name: str) -> bool:
        return all(reader.has_reference(name) for reader in self._readers)

    def close(self) -> None:
        for f in self._readers:
            f.close()



from whatshap.core import Read


def verify_mec_score_and_partitioning(dp_table, reads):
    """Confirms that the results reported by dp_table are consistent: check
    whether the reported partitioning leads to the reported MEC score."""
    superreads, transmission_vector = dp_table.get_super_reads()
    assert len(superreads) == 1
    superreads = superreads[0]
    assert len(superreads) == 2
    # create new superreads that don't contain 3s (EQUAL COST)
    new_superreads = [Read("superread0", 0), Read("superread1", 0)]
    assert len(superreads[0]) == len(superreads[1])
    for i in range(len(superreads[0])):
        for j in range(2):
            v = superreads[j][i]
            allele = v.allele
            if allele == 3:
                allele = j
            new_superreads[j].add_variant(v.position, allele, v.quality)
    partitioning = dp_table.get_optimal_partitioning()
    position_to_index = {variant.position: index for index, variant in enumerate(new_superreads[0])}
    swapped = False
    mec_score = 0
    n = 0
    for read_index, read in enumerate(reads):
        cost0 = 0
        cost1 = 0
        for variant in read:
            if variant.position in position_to_index:
                if new_superreads[0][position_to_index[variant.position]].allele != variant.allele:
                    cost0 = cost0 + variant.quality
                if new_superreads[1][position_to_index[variant.position]].allele != variant.allele:
                    cost1 = cost1 + variant.quality
        mec_score += min(cost0, cost1)
        if cost0 == cost1:
            continue
        haplotype = 0 if (cost0 < cost1) != swapped else 1
        if partitioning[read_index] != haplotype:
            if n == 0:
                swapped = True
            else:
                assert False
        n += 1
    print(
        "Expected MEC score: {}, obtained MEC score: {}".format(
            mec_score, dp_table.get_optimal_cost()
        )
    )
    assert mec_score == dp_table.get_optimal_cost()



from xopen import xopen
import pysam

from whatshap.cli.split import run_split


def test_split_bam(tmp_path):
    h1 = tmp_path / "h1.bam"
    h2 = tmp_path / "h2.bam"

    run_split(
        "tests/data/pacbio/pacbio.bam",
        "tests/data/pacbio/haplotags.txt",
        output_h1=h1,
        output_h2=h2,
    )
    with pysam.AlignmentFile(h1) as f:
        assert 15 == len(list(f))
    with pysam.AlignmentFile(h2) as f:
        assert 10 == len(list(f))


def test_split_bam_no_sequence(tmp_path):
    """
    Test that a BAM file w/o sequence records
    can be processed - see issue 215
    """

    expected_output = [
        "205\t1\t0\t0\n",
        "716\t1\t0\t0\n",
        "1613\t0\t0\t1\n",
        "2250\t1\t0\t0\n",
        "3551\t1\t0\t0\n",
        "4385\t1\t0\t0\n",
        "6750\t1\t0\t0\n",
        "11263\t1\t0\t0\n",
        "12930\t0\t1\t0\n",
        "23225\t0\t1\t0\n",
    ]
    rlen_hist = tmp_path / "rlenhist.tsv"
    input_bam = "tests/data/reads-no-sequence.bam"
    input_list = "tests/data/reads-no-sequence.haplotags.tsv"
    run_split(
        input_bam,
        input_list,
        output_h1="/dev/null",
        output_h2="/dev/null",
        output_untagged="/dev/null",
        read_lengths_histogram=rlen_hist,
    )

    with open(rlen_hist, "r") as dump:
        produced_output = dump.readlines()[1:]  # skip header line
        for e, p in zip(expected_output, produced_output):
            assert e == p


def test_split_fastq(tmp_path):
    # issue #371
    fastq_path = tmp_path / "reads.fastq.gz"
    list_path = tmp_path / "readlist.txt"
    with xopen(fastq_path, "w") as f:
        f.write("@r\nACGT\n+\n####\n")
    list_path.write_text("hello\tH1")
    run_split(
        str(fastq_path),
        str(list_path),
        output_h1="/dev/null",
        output_h2="/dev/null",
    )



"""
Test genotyping of pedigrees
"""
import math

from whatshap.core import (
    GenotypeDPTable,
    ReadSet,
    Pedigree,
    NumericSampleIds,
    PhredGenotypeLikelihoods,
    Genotype,
)
from whatshap.testhelpers import (
    string_to_readset_pedigree,
    canonic_index_list_to_biallelic_gt_list,
)


def genotype_pedigree(
    numeric_sample_ids,
    reads,
    recombcost,
    pedigree,
    expected_genotypes,
    weights=None,
    expected=None,
    scaling=10,
    positions=None,
):
    rs = string_to_readset_pedigree(s=reads, w=weights, scaling_quality=scaling)
    dp_forward_backward = GenotypeDPTable(numeric_sample_ids, rs, recombcost, pedigree, positions)

    # for each position compare the likeliest genotype to the expected ones
    print("expected genotypes: ", expected_genotypes)
    positions = rs.get_positions()
    for pos in range(len(positions)):
        for individual in range(len(pedigree)):
            likelihoods = dp_forward_backward.get_genotype_likelihoods(
                "individual" + str(individual), pos
            )

            # if expected likelihoods given, compare
            if expected is not None:
                print(
                    "likelihoods: ",
                    likelihoods,
                    " expected likelihoods: ",
                    expected[individual][pos],
                )
                assert likelihoods == PhredGenotypeLikelihoods(expected[individual][pos])

            # find the likeliest genotype
            max_val = -1
            max_geno = Genotype([])
            for genotype in likelihoods.genotypes():
                assert not math.isnan(likelihoods[genotype])
                if likelihoods[genotype] > max_val:
                    max_val = likelihoods[genotype]
                    max_geno = genotype

            # compare it to the expected genotype
            print(
                "pos.: " + str(pos) + " individual " + str(individual) + ": ",
                likelihoods,
                " expected genotype: ",
                expected_genotypes[individual][pos],
            )
            assert max_geno == expected_genotypes[individual][pos]
        print("\n")


def test_genotyping_empty_trio():
    rs = ReadSet()
    recombcost = []
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual("individual0", [], [])
    pedigree.add_individual("individual1", [], [])
    pedigree.add_individual("individual2", [], [])
    pedigree.add_relationship("individual0", "individual1", "individual2")
    _ = GenotypeDPTable(numeric_sample_ids, rs, recombcost, pedigree)


def test_genotyping_trio1():
    reads = """
      A 00
      A 00
      B 11
      B 11
      C 11
      C 00
    """

    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([0, 0]),
        canonic_index_list_to_biallelic_gt_list([2, 2]),
        canonic_index_list_to_biallelic_gt_list([1, 1]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([1, 1]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 2,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([1, 1]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 2,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([1, 1]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 2,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [10, 10]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


def test_genotyping_quartet1():
    reads = """
         A 1111
         A 0000
         B 1010
         C 111000
         C 010101
         D 000000
         D 010
         B 0101
         C  1100
         D  10010
         A   0000
         A   1111
         B   1010
         B   0101
       """
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([1, 2, 1, 1, 0, 1]),
        canonic_index_list_to_biallelic_gt_list([0, 1, 0, 0, 1, 0]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual3",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    pedigree.add_relationship("individual0", "individual1", "individual3")
    recombcost = [3, 3, 3, 4, 3, 3]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


def test_genotyping_trio2():
    reads = """
      A 00
      A 00
      B 11
      B 11
      C 11
      C 00
    """
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([0, 0]),
        canonic_index_list_to_biallelic_gt_list([2, 2]),
        canonic_index_list_to_biallelic_gt_list([1, 1]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 2,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 2,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 2,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [10, 10, 10]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


def test_genotyping_trio3():
    reads = """
      A 1111
      B 1010
      C 111000
      C 010101
      C 010101
      B 0101
      A  0000
      B  1010
      C  1010
      C  1100
      A   0000
      A   1111
      B   1010
      B    010
    """
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([1, 2, 1, 1, 0, 1]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [3, 3, 3, 4, 3, 3]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


# TODO: what about such cases, where the given reads are like below? What would be the expected genotypes
# when genotyping with higher recombrate (e.g. 10), resulting genotypes are: [2,2,2] , [2,1,1], [2,2,2], why??
def test_genotyping_trio4():
    reads = """
      B 101
      B 101
      B 101
      A 111
      A 111
      A 111
      C 111
      C 111
      C 111
    """
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([2, 2, 2]),
        canonic_index_list_to_biallelic_gt_list([2, 1, 2]),
        canonic_index_list_to_biallelic_gt_list([2, 2, 2]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [1, 1, 1]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


def test_genotyping_trio5():
    reads = """
      B 101
      B 101
      B 101
      A 111
      A 111
      A 111
      C 111
      C 111
      C 101
      C 101
    """
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([2, 2, 2]),
        canonic_index_list_to_biallelic_gt_list([2, 0, 2]),
        canonic_index_list_to_biallelic_gt_list([2, 1, 2]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [2, 2, 2]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


def test_genotyping_trio6():
    reads = """
      A 000
      A 000
      A 010
      A 111
      A 111
      B 111
      B 111
      C 111
      C 000
      C 000
    """

    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([2, 2, 2]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 1]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([1, 1, 1]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([1, 1, 1]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([1, 1, 1]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [10, 10, 10]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


def test_genotyping_quartet2():
    reads = """
      A 111
      A 010
      A 110
      B 001
      B 110
      B 101
      C 001
      C 010
      C 010
      D 001
      D 010
      D 010
    """
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([1, 2, 0]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([0, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([0, 1, 1]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_individual(
        "individual3",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    pedigree.add_relationship("individual0", "individual1", "individual3")
    recombcost = [10, 10, 10]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


def test_genotyping_quartet3():
    reads = """
      A 111111
      A 000000
      B 010101
      B 101010
      C 000000
      C 010101
      D 000000
      D 010101
    """
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([0, 1, 0, 1, 0, 1]),
        canonic_index_list_to_biallelic_gt_list([0, 1, 0, 1, 0, 1]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual3",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    pedigree.add_relationship("individual0", "individual1", "individual3")
    recombcost = [3, 3, 3, 3, 3, 3]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


def test_genotyping_quartet4():
    reads = """
      A 1111
      A 0000
      B 1010
      C 111000
      C 010101
      D 000000
      D 010
      B 0101
      C  1100
      D  10010
      A   0000
      A   1111
      B   1010
      B   0101
    """
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([1, 2, 1, 1, 0, 1]),
        canonic_index_list_to_biallelic_gt_list([0, 1, 0, 0, 1, 0]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual3",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 6,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    pedigree.add_relationship("individual0", "individual1", "individual3")
    recombcost = [3, 3, 3, 4, 3, 3]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


def test_genotyping_trio7():
    reads = """
      B 100
      B 100
      B 111
      A 111
      A 111
      A 111
      C 111
      C 101
      C 101
    """
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([2, 2, 2]),
        canonic_index_list_to_biallelic_gt_list([2, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([2, 1, 2]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 3,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [1, 1, 1]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


def test_genotyping_trio8():
    reads = """
      B 1100
      B 1110
      A 1111
      A 0000
      C 0011
      C 1110
    """
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([2, 2, 1, 0]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 2, 1]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 4,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 4,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 4,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [10, 10, 10, 10]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


def test_genotyping_trio9():
    reads = """
      B 1100
      B 1100
      B 1100
      B 1110
      B 1110
      B 1110
      A 1111
      A 1111
      A 1111
      A 0000
      A 0000
      A 0000
      C 0011
      C 0011
      C 1110
      C 1110
    """
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([2, 2, 1, 0]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 2, 1]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 4,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 4,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 4,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [10, 10, 10, 10]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


# TODO when using uniform priors (1/3,1/3,1/3) result for child is (0,0.2,0.8)
def test_weighted_genotyping():
    reads = """
      B 00
      B 11
      A 11
      A 00
      C 11
      C 11
    """
    weights = """
      99
      99
      99
      99
      99
      99
    """
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([1, 1]),
        canonic_index_list_to_biallelic_gt_list([1, 1]),
        canonic_index_list_to_biallelic_gt_list([2, 2]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([0.25, 0.5, 0.25])] * 4,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([0.25, 0.5, 0.25])] * 4,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([0.25, 0.5, 0.25])] * 4,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    # recombination is extremely unlikely
    recombcost = [1000, 1000, 1000, 1000]

    expected = {
        0: [[0, 1, 0], [0, 1, 0]],
        1: [[0, 1, 0], [0, 1, 0]],
        2: [[0, 1.0 / 3.0, 2 / 3.0], [0, 1.0 / 3.0, 2 / 3.0]],
    }
    genotype_pedigree(
        numeric_sample_ids,
        reads,
        recombcost,
        pedigree,
        expected_genotypes,
        weights,
        expected,
        scaling=500,
    )


def test_genotyping_trio10():
    reads = """
      B 0000
      B 0000
      B 0000
      B 0000
      B 0000
      B 0000
      A 1111
      A 1111
      A 1111
      A 1111
      A 1111
      A 1111
    """

    # no reads for child, but genotype must be 1/0 for each pos. (due to inheritance)
    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([2, 2, 2, 2]),
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 4,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 4,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 4,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [10, 10, 10, 10]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


def test_genotyping_trio11():
    reads = """
      A 111
      B 110
      B 111
      C 000
      C 110
    """

    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([2, 2, 1]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 0]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 4,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 4,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * 4,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [10, 10, 10]
    genotype_pedigree(numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes)


# TODO: model fails to infer the correct genotype likelihoods of the child, if uniform priors are used for child.
# according to mendelian inheritance, correct priors here would be:
# A: (0,1,0), B:(0,1,0), C:(0.25,0.5,0.25), but since no reads present for child,
# prior genotyping step would give uniform priors.
def test_genotyping_trio13():
    reads = """
      A 1111
      A 0000
      B 1111
      B 0000
    """

    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([0, 1, 0])] * 6,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([0, 1, 0])] * 6,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([0.25, 0.5, 0.25])] * 6,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [1000000, 1000000, 1000000, 1000000, 1000000, 1000000]
    genotype_pedigree(
        numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes, scaling=1000
    )


def test_genotyping_trio14():
    reads = """
      A 111111
      A 111111
      B 111111
      B 000000
      C 000000
    """

    expected_genotypes = [
        canonic_index_list_to_biallelic_gt_list([2, 2, 2, 2, 2, 2]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
        canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1]),
    ]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1 / 3.0, 1 / 3.0, 1 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual1",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1 / 3.0, 1 / 3.0, 1 / 3.0])] * 6,
    )
    pedigree.add_individual(
        "individual2",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0, 0, 0, 0]),
        [PhredGenotypeLikelihoods([1 / 3.0, 1 / 3.0, 1 / 3.0])] * 6,
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [1000000, 1000000, 1000000, 1000000, 1000000, 1000000]
    genotype_pedigree(
        numeric_sample_ids, reads, recombcost, pedigree, expected_genotypes, scaling=1000
    )



from typing import Tuple, Optional, Union

Item = int

class PriorityQueue:
    def push(self, score: int, item: Item) -> None: ...
    def pop(self) -> Tuple[int, Item]: ...
    def change_score(self, item: Item, new_score: int) -> None: ...
    def get_score_by_item(self, item: Item) -> Optional[Union[Tuple, int]]: ...
    def __len__(self) -> int: ...
    def is_empty(self) -> bool: ...



"""
Integration tests that use the command-line entry point run_polyphase.
"""
import os

from pytest import raises
from whatshap.cli.polyphase import run_polyphase, CommandLineError
from whatshap.vcf import VcfReader


def test_polyphase_short_chr22(tmp_path):
    outvcf = tmp_path / "output.vcf"
    run_polyphase(
        phase_input_files=["tests/data/polyploid.chr22.42M.12k.bam"],
        variant_file="tests/data/polyploid.chr22.42M.12k.vcf",
        ploidy=4,
        ignore_read_groups=True,
        output=outvcf,
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "chr22"
    assert len(table.variants) == 42
    assert table.samples == ["HG00514_NA19240"]


def test_polyphase_multiple_bam(tmp_path):
    outvcf = tmp_path / "output.vcf"
    run_polyphase(
        phase_input_files=[
            "tests/data/polyploid.human1.chr22.42M.5k.bam",
            "tests/data/polyploid.human2.chr22.42M.5k.bam",
        ],
        variant_file="tests/data/polyploid.multisample.chr22.42M.5k.vcf",
        ploidy=2,
        ignore_read_groups=False,
        output=outvcf,
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "chr22"
    assert len(table.variants) == 9
    assert set(table.samples) == set(["HG00514", "NA19240"])
    assert not all(p is None for p in table.phases_of("HG00514"))
    assert not all(p is None for p in table.phases_of("NA19240"))


def test_wrong_ploidy(tmp_path):
    outvcf = tmp_path / "output.vcf"
    with raises(CommandLineError):
        run_polyphase(
            phase_input_files=["tests/data/polyploid.chr22.42M.12k.bam"],
            variant_file="tests/data/polyploid.chr22.42M.12k.vcf",
            ploidy=3,
            ignore_read_groups=True,
            output=outvcf,
        )


def test_blockcut_sensitivities(tmp_path):
    """Ensure that the block cut sets are monotone to the sensitivity"""

    results = []
    for s in range(6):
        outvcf = tmp_path / "output{}.vcf".format(s)
        run_polyphase(
            phase_input_files=["tests/data/polyploid.chr22.42M.12k.bam"],
            variant_file="tests/data/polyploid.chr22.42M.12k.vcf",
            ploidy=4,
            ignore_read_groups=True,
            block_cut_sensitivity=s,
            output=outvcf,
        )
        assert os.path.isfile(outvcf)

        tables = list(VcfReader(outvcf, phases=True))
        assert len(tables) == 1
        block_starts = set(
            [i.block_id for i in tables[0].phases_of("HG00514_NA19240") if i is not None]
        )
        results.append(block_starts)
        print(block_starts)

    for s in range(5):
        assert all(cut in results[s + 1] for cut in results[s])


def test_polyphase_multithreaded(tmp_path):
    outvcf_st = tmp_path / "output_st.vcf"
    outvcf_mt = tmp_path / "output_mt.vcf"

    run_polyphase(
        phase_input_files=["tests/data/polyploid.chr22.42M.12k.bam"],
        variant_file="tests/data/polyploid.chr22.42M.12k.vcf",
        ploidy=4,
        ignore_read_groups=True,
        output=outvcf_st,
    )
    run_polyphase(
        phase_input_files=["tests/data/polyploid.chr22.42M.12k.bam"],
        variant_file="tests/data/polyploid.chr22.42M.12k.vcf",
        ploidy=4,
        ignore_read_groups=True,
        output=outvcf_mt,
        threads=4,
    )
    assert os.path.isfile(outvcf_st)
    assert os.path.isfile(outvcf_mt)

    tables_st = list(VcfReader(outvcf_st, phases=True))
    table_st = tables_st[0]
    tables_mt = list(VcfReader(outvcf_mt, phases=True))
    table_mt = tables_mt[0]

    assert table_st.chromosome == table_mt.chromosome
    assert table_st.samples == table_mt.samples
    assert all([st == mt for (st, mt) in zip(table_st.genotypes, table_mt.genotypes)])
    assert all([st == mt for (st, mt) in zip(table_st.phases, table_mt.phases)])
    assert all(
        [st == mt for (st, mt) in zip(table_st.genotype_likelihoods, table_mt.genotype_likelihoods)]
    )
    assert all([st == mt for (st, mt) in zip(table_st.variants, table_mt.variants)])
